# general imports
import jax
import jax.numpy as jnp
from functools import partial

# typing imports
from typing import Union
from beartype import beartype as typechecker
from jaxtyping import Array, Float, jaxtyped

# general astronomix imports
from astronomix._physics_modules._cosmic_rays.cr_fluid_equations import speed_of_sound_crs
from astronomix._physics_modules._self_gravity._poisson_solver import (
    _compute_gravitational_potential,
)
from astronomix._finite_volume._state_evolution.limiters import _van_albada_limiter, _minmod
from astronomix._stencil_operations._stencil_operations import _stencil_add
from astronomix.option_classes.simulation_config import (
    CARTESIAN,
    MUSCL,
    STATE_TYPE,
    STATE_TYPE_ALTERED,
    VAN_ALBADA,
    VAN_ALBADA_PP,
    SimulationConfig,
)
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.variable_registry.registered_variables import RegisteredVariables

# speed of sound calculation
from astronomix._fluid_equations._equations import speed_of_sound

# limited gradients
from astronomix._finite_volume._state_evolution.limited_gradients import _calculate_limited_gradients
from astronomix.option_classes.simulation_params import SimulationParams


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables", "axis"])
def _reconstruct_at_interface_split(
    primitive_state: STATE_TYPE,
    dt: Union[float, Float[Array, ""]],
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
    axis: int,
) -> tuple[STATE_TYPE_ALTERED, STATE_TYPE_ALTERED]:

    # get fluid variables for convenience
    rho = primitive_state[registered_variables.density_index]
    p = primitive_state[registered_variables.pressure_index]
    u = primitive_state[axis]

    # get the limited gradients on the cells
    limited_gradients = _calculate_limited_gradients(
        primitive_state, config, helper_data, axis=axis
    )

    if config.time_integrator == MUSCL:
        # calculate the sound speed
        if not config.cosmic_ray_config.cosmic_rays:
            c = speed_of_sound(rho, p, gamma)
        else:
            c = speed_of_sound_crs(primitive_state, registered_variables)

        # ================ construct A_W, the "primitive Jacabian" (not an actual Jacabian) ================
        # see https://diglib.uibk.ac.at/download/pdf/4422963.pdf, 2.11

        # calculate the vectors making up A_W
        A_W = jnp.zeros((registered_variables.num_vars,) + primitive_state.shape)

        # set u diagonal, this way all quantities are automatically advected
        A_W = A_W.at[
            jnp.arange(registered_variables.num_vars),
            jnp.arange(registered_variables.num_vars),
        ].set(u)

        # set rest
        A_W = A_W.at[registered_variables.density_index, axis].set(rho)
        A_W = A_W.at[registered_variables.pressure_index, 1].set(rho * c**2)
        A_W = A_W.at[axis, registered_variables.pressure_index].set(1 / rho)

        # ====================================================================================================

        # project the gradients
        if config.dimensionality == 1:
            projected_gradients = jnp.einsum("bax, ax -> bx", A_W, limited_gradients)
        elif config.dimensionality == 2:
            projected_gradients = jnp.einsum("baxy, axy -> bxy", A_W, limited_gradients)
        elif config.dimensionality == 3:
            projected_gradients = jnp.einsum(
                "baxyz, axyz -> bxyz", A_W, limited_gradients
            )

        # predictor step
        predictors = primitive_state - dt / 2 * projected_gradients
    else:
        raise ValueError(
            f"Time integrator {config.time_integrator} not supported for split reconstruction. Only MUSCL is supported."
        )

    # compute primitives at the interfaces
    if config.geometry == CARTESIAN:
        distances_to_left_interfaces = (
            config.grid_spacing / 2
        )  # distances r_i - r_{i-1/2}
        distances_to_right_interfaces = (
            config.grid_spacing / 2
        )  # distances r_{i+1/2} - r_i
    else:
        r = helper_data.geometric_centers
        rv = helper_data.volumetric_centers

        distances_to_left_interfaces = rv - (r - config.grid_spacing / 2)
        distances_to_right_interfaces = (r + config.grid_spacing / 2) - rv

    primitives_left = predictors - distances_to_left_interfaces * limited_gradients
    primitives_right = predictors + distances_to_right_interfaces * limited_gradients

    # primitives left at i is the left state at the interface
    # between i-1 and i so the right extrapolation from the cell i-1
    p_left_interface = jnp.roll(primitives_right, shift=1, axis=axis)

    # primitives right at i is the right state at the interface
    # between i-1 and i so the left extrapolation from the cell i
    p_right_interface = primitives_left

    return p_left_interface, p_right_interface


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _reconstruct_at_interface_unsplit(
    primitive_state: STATE_TYPE,
    dt: Union[float, Float[Array, ""]],
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    params: SimulationParams,
    helper_data: HelperData,
    registered_variables: RegisteredVariables,
):
    """
    Unsplit reconstruction.
    """

    # this is very memory inefficient!!!

    # limited gradients: dimensionality x state_shape
    limited_gradients = jnp.zeros((config.dimensionality,) + primitive_state.shape)

    for axis in range(1, config.dimensionality + 1):
        limited_gradients = limited_gradients.at[axis - 1].set(
            _calculate_limited_gradients(
                primitive_state, config, helper_data, axis=axis
            )
        )

    differences = limited_gradients * config.grid_spacing / 2

    primitives_left_interface = jnp.zeros(
        (config.dimensionality,) + primitive_state.shape
    )
    primitives_right_interface = jnp.zeros(
        (config.dimensionality,) + primitive_state.shape
    )

    if config.limiter == VAN_ALBADA_PP:
        # positivity preserving reconstruction

        eps = 1e-14

        rho = primitive_state[registered_variables.density_index]
        p = primitive_state[registered_variables.pressure_index]
        c = speed_of_sound(rho, p, gamma)
        alpha_lax = jnp.zeros((config.dimensionality,))
        for axis in range(1, config.dimensionality + 1):
            u = primitive_state[axis]
            alpha_lax_i = jnp.max(jnp.abs(u) + c)
            alpha_lax = alpha_lax.at[axis - 1].set(alpha_lax_i)

        # NOTE: formula will change for different grid spacings along dimensions!!!

        C = alpha_lax / jnp.sum(alpha_lax)

        if config.mhd:
            # in the MHD case we go half-time steps as of
            # the strang splitting, so effectively, for the
            # hydro part we use C_cfl / 2, so 1 / (C_cfl / 2)
            # = 2 / C_cfl
            q = 2 / params.C_cfl
        else:
            q = 1 / params.C_cfl

        density_diff_protected = jnp.where(
            jnp.abs(differences[:, registered_variables.density_index]) > eps,
            differences[:, registered_variables.density_index],
            eps,
        )

        pressure_diff_protected = jnp.where(
            jnp.abs(differences[:, registered_variables.pressure_index]) > eps,
            differences[:, registered_variables.pressure_index],
            eps,
        )

        alpha_density = jnp.where(
            jnp.abs(differences[:, registered_variables.density_index]) > eps,
            jnp.minimum(
                primitive_state[registered_variables.density_index]
                / (jnp.abs(density_diff_protected) * (1 + eps)),
                1,
            ),
            1,
        )

        kappa_pressure = jnp.where(
            jnp.abs(differences[:, registered_variables.pressure_index]) > eps,
            jnp.minimum(
                primitive_state[registered_variables.pressure_index]
                / (jnp.abs(pressure_diff_protected) * (1 + eps)),
                1,
            ),
            1,
        )

        if config.dimensionality == 1:
            A1 = jnp.sum(
                jnp.sum(
                    C[:, None]
                    * alpha_density
                    * differences[:, registered_variables.density_index]
                    * differences[:, registered_variables.velocity_index],
                    axis=0,
                )
                ** 2,
                axis=0,
            )
            A2 = jnp.sum(
                C[:, None]
                * jnp.sum(
                    differences[:, registered_variables.velocity_index] ** 2, axis=1
                ),
                axis=0,
            )
        elif config.dimensionality == 2:
            A1 = jnp.sum(
                jnp.sum(
                    C[:, None, None]
                    * alpha_density
                    * differences[:, registered_variables.density_index]
                    * differences[
                        :,
                        registered_variables.velocity_index.x : registered_variables.velocity_index.x
                        + config.dimensionality,
                    ],
                    axis=0,
                )
                ** 2,
                axis=0,
            )
            A2 = jnp.sum(
                C[:, None, None]
                * jnp.sum(
                    differences[
                        :,
                        registered_variables.velocity_index.x : registered_variables.velocity_index.x
                        + config.dimensionality,
                    ]
                    ** 2,
                    axis=1,
                ),
                axis=0,
            )
        elif config.dimensionality == 3:
            A1 = jnp.sum(
                jnp.sum(
                    C[:, None, None, None]
                    * alpha_density
                    * differences[:, registered_variables.density_index]
                    * differences[
                        :,
                        registered_variables.velocity_index.x : registered_variables.velocity_index.x
                        + config.dimensionality,
                    ],
                    axis=0,
                )
                ** 2,
                axis=0,
            )
            A2 = jnp.sum(
                C[:, None, None, None]
                * jnp.sum(
                    differences[
                        :,
                        registered_variables.velocity_index.x : registered_variables.velocity_index.x
                        + config.dimensionality,
                    ]
                    ** 2,
                    axis=1,
                ),
                axis=0,
            )
        vsum = jnp.sum(
            jnp.sum(
                differences[
                    :,
                    registered_variables.velocity_index.x : registered_variables.velocity_index.x
                    + config.dimensionality,
                ]
                ** 2,
                axis=1,
            ),
            axis=0,
        )
        A1 = jnp.where(vsum > eps, A1, eps)
        A2 = jnp.where(vsum > eps, A2, eps)

        beta = jnp.where(
            vsum > eps,
            jnp.minimum(
                jnp.sqrt(
                    (
                        (q - 2) ** 2
                        * primitive_state[registered_variables.density_index]
                        * primitive_state[registered_variables.pressure_index]
                    )
                    / (
                        (gamma - 1)
                        * (
                            2 * A1
                            + (q - 2)
                            * primitive_state[registered_variables.density_index] ** 2
                            * A2
                        )
                    )
                ),
                1,
            ),
            1,
        )

        differences_pp = differences

        differences_pp = differences_pp.at[:, registered_variables.density_index].set(
            differences[:, registered_variables.density_index] * alpha_density
        )

        differences_pp = differences_pp.at[:, registered_variables.pressure_index].set(
            differences[:, registered_variables.pressure_index] * kappa_pressure
        )

        differences_pp = differences_pp.at[
            :,
            registered_variables.velocity_index.x : registered_variables.velocity_index.x
            + config.dimensionality,
        ].set(
            differences[
                :,
                registered_variables.velocity_index.x : registered_variables.velocity_index.x
                + config.dimensionality,
            ]
            * beta
        )

        differences = differences_pp

    for axis in range(1, config.dimensionality + 1):
        # i-1/2R, ...
        primitives_left_center = (
            primitive_state - differences[axis - 1]
        )  # left of the cell center but the right of the interface
        # i+1/2L, ...
        primitives_right_center = (
            primitive_state + differences[axis - 1]
        )  # right of the cell center but the left of the interface

        # primitives left at i is the left state at the interface
        # between i-1 and i so the right extrapolation from the cell i-1
        p_left_interface = jnp.roll(primitives_right_center, shift=1, axis=axis)

        # primitives right at i is the right state at the interface
        # between i-1 and i so the left extrapolation from the cell i
        p_right_interface = primitives_left_center

        # set the values
        primitives_left_interface = primitives_left_interface.at[axis - 1].set(
            p_left_interface
        )
        primitives_right_interface = primitives_right_interface.at[axis - 1].set(
            p_right_interface
        )

    return primitives_left_interface, primitives_right_interface


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "axis"])
def _reconstruct_at_interface_unsplit_single(
    primitive_state: STATE_TYPE,
    config: SimulationConfig,
    helper_data: HelperData,
    axis: int,
):
    """
    Unsplit reconstruction.
    """

    limited_gradients = _calculate_limited_gradients(
        primitive_state, config, helper_data, axis=axis
    )
    differences = limited_gradients * config.grid_spacing / 2

    # i-1/2R, ...
    primitives_left_center = (
        primitive_state - differences
    )  # left of the cell center but the right of the interface
    # i+1/2L, ...
    primitives_right_center = (
        primitive_state + differences
    )  # right of the cell center but the left of the interface

    # primitives left at i is the left state at the interface
    # between i-1 and i so the right extrapolation from the cell i-1
    p_left_interface = jnp.roll(primitives_right_center, shift=1, axis=axis)

    # primitives right at i is the right state at the interface
    # between i-1 and i so the left extrapolation from the cell i
    p_right_interface = primitives_left_center

    return p_left_interface, p_right_interface



import jax
import jax.numpy as jnp
import equinox as eqx
from jaxtyping import Array, Float, PRNGKeyArray

# general
from functools import partial
import jax.numpy as jnp
import jax

import equinox as eqx

# typing
from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker
from typing import Tuple, Union

from astronomix._finite_volume._magnetic_update._vector_maths import curl2D
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import STATE_TYPE, SimulationConfig
from astronomix.option_classes.simulation_params import SimulationParams


class CorrectorCNN(eqx.Module):
    """
    A simple CNN that maps an input of shape (C, H, W) to an output of the same shape.
    """

    layers: list

    def __init__(self, in_channels: int, hidden_channels: int, *, key: PRNGKeyArray):
        # We need a key for each convolutional layer
        key1, key2, key3 = jax.random.split(key, 3)

        # A simple 3-layer CNN.
        # Note the use of padding=1 with kernel_size=3 to keep the
        # spatial dimensions (height and width) the same.
        self.layers = (
            # Layer 1: Expand channels from NUM_VARS to HIDDEN_CHANNELS
            eqx.nn.Conv2d(
                in_channels, hidden_channels, kernel_size=3, padding=1, key=key1
            ),
            jax.nn.relu,
            # Layer 2: A hidden convolutional layer
            eqx.nn.Conv2d(
                hidden_channels, hidden_channels, kernel_size=3, padding=1, key=key2
            ),
            jax.nn.relu,
            # Layer 3: Contract channels back to the original NUM_VARS
            # No activation function here, as we want to predict a raw correction value.
            eqx.nn.Conv2d(
                hidden_channels, in_channels, kernel_size=3, padding=1, key=key3
            ),
        )

    def __call__(self, x: Float[Array, "num_vars h w"]) -> Float[Array, "num_vars h w"]:
        """
        The forward pass of the model.
        """
        # Pass the input through the network to get the correction term
        correction = x
        for layer in self.layers:
            correction = layer(correction)

        # Add the learned correction to the original input
        return correction


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["registered_variables", "config"])
def _cnn_mhd_corrector(
    primitive_state: STATE_TYPE,
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    params: SimulationParams,
    time_step: Float[Array, ""],
):
    neural_net_params = params.cnn_mhd_corrector_params.network_params
    neural_net_static = config.cnn_mhd_corrector_config.network_static
    model = eqx.combine(neural_net_params, neural_net_static)

    correction = model(primitive_state)

    # to not add divergence errors, we learn a correction for the electric field
    # - and the divergence of a curl is zero
    electric_field_correction = correction[-3:, ...]
    magnetic_field_correction = curl2D(electric_field_correction, config.grid_spacing)
    correction = correction.at[-3:, ...].set(magnetic_field_correction)

    # update the primitive state with the correction
    primitive_state = primitive_state + correction * time_step

    # ensure that the pressure is larger than a minimum value
    p_min = 1e-4
    primitive_state = primitive_state.at[registered_variables.pressure_index].set(
        jnp.maximum(primitive_state[registered_variables.pressure_index], p_min)
    )

    return primitive_state



# data structures
from astronomix.option_classes.simulation_config import SimulationConfig
from astronomix.option_classes.simulation_params import SimulationParams

# constants
from astronomix.option_classes.simulation_config import (
    FORWARDS,
    BACKWARDS,
    MINMOD,
    OSHER,
    HLL,
    HLLC,
    HLLC_LM,
    OPEN_BOUNDARY,
    REFLECTIVE_BOUNDARY,
    PERIODIC_BOUNDARY,
    CARTESIAN,
    CYLINDRICAL,
    SPHERICAL,
)

# initialization functions
from astronomix.data_classes.simulation_helper_data import get_helper_data
from astronomix.variable_registry.registered_variables import get_registered_variables
from astronomix.option_classes.simulation_config import finalize_config
from astronomix.initial_condition_generation.construct_primitive_state import construct_primitive_state
from astronomix._finite_difference._magnetic_update._constrained_transport import (
    initialize_interface_fields,
)

# module-setup
from astronomix._physics_modules._stellar_wind.stellar_wind_options import WindParams

# run
from astronomix.time_stepping.time_integration import time_integration

# units
from astronomix.units import CodeUnits



"""
Fourier-based Poisson solver and simple source term handling
of self gravity. To be improved to an energy-conserving scheme.
"""

# general
from functools import partial
import jax.numpy as jnp
import jax

# typing
from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker
from typing import Tuple, Union

# astronomix data classes
from astronomix._physics_modules._cosmic_rays.cr_fluid_equations import speed_of_sound_crs
from astronomix._physics_modules._self_gravity._poisson_solver import (
    _compute_gravitational_potential,
)
from astronomix._finite_volume._riemann_solver._riemann_solver import _riemann_solver
from astronomix._stencil_operations._stencil_operations import _stencil_add
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import (
    DONOR_ACCOUNTING,
    GHOST_CELLS,
    HALF_SPLIT,
    HLLC_LM,
    LAX_FRIEDRICHS,
    RIEMANN_SPLIT,
    RIEMANN_SPLIT_UNSTABLE,
    SIMPLE_SOURCE_TERM,
    SPLIT,
    STATE_TYPE_ALTERED,
    SimulationConfig,
)

# astronomix constants
from astronomix.option_classes.simulation_config import (
    FIELD_TYPE,
    HLL,
    HLLC,
    OPEN_BOUNDARY,
    STATE_TYPE,
)

# astronomix functions
from astronomix._geometry.boundaries import _boundary_handler
from astronomix._finite_volume._riemann_solver.hll import _hll_solver, _hllc_solver
from astronomix._finite_volume._state_evolution.reconstruction import (
    _reconstruct_at_interface_split,
    _reconstruct_at_interface_unsplit,
)
from astronomix._fluid_equations._fluxes import _euler_flux
from astronomix._fluid_equations._equations import (
    conserved_state_from_primitive,
    primitive_state_from_conserved,
    speed_of_sound,
)
from astronomix.option_classes.simulation_params import SimulationParams


# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit, static_argnames=["axis", "grid_spacing", "registered_variables", "config"]
)
def _gravitational_source_term_along_axis(
    gravitational_potential: FIELD_TYPE,
    primitive_state: STATE_TYPE,
    grid_spacing: float,
    registered_variables: RegisteredVariables,
    dt: Union[float, Float[Array, ""]],
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    params: SimulationParams,
    helper_data: HelperData,
    axis: int,
) -> STATE_TYPE:
    """
    Compute the source term for the self-gravity solver along a single axis.
    Currently, simply density * gravitational_acceleration for the momentum
    and density * velocity * gravitational_acceleration for the energy.

    Args:
        gravitational_potential: The gravitational potential.
        primitive_state: The primitive state.
        grid_spacing: The grid spacing.
        registered_variables: The registered variables.
        dt: The time step.
        gamma: The adiabatic index.
        config: The simulation configuration.
        helper_data: The helper data.
        axis: The axis along which to compute the source term.

    Returns:
        The source term.

    """

    rho = primitive_state[registered_variables.density_index]
    v_axis = primitive_state[axis]

    # a_i = - (phi_{i+1} - phi_{i-1}) / (2 * dx)
    acceleration = -_stencil_add(
        gravitational_potential, indices=(1, -1), factors=(1.0, -1.0), axis=axis - 1
    ) / (2 * grid_spacing)
    # it is axis - 1 because the axis is 1-indexed as usually the zeroth axis are the different
    # fields in the state vector not the spatial dimensions, but here we only have the spatial dimensions

    source_term = jnp.zeros_like(primitive_state)

    # set momentum source
    source_term = source_term.at[axis].set(rho * acceleration)

    if config.self_gravity_version == SIMPLE_SOURCE_TERM:
        # set energy source
        source_term = source_term.at[registered_variables.pressure_index].set(
            rho * v_axis * acceleration
        )

    else:
        # the other schemes are based on the reconstructed states
        if config.first_order_fallback:
            primitive_state_left = jnp.roll(primitive_state, shift=1, axis=axis)
            primitive_state_right = primitive_state
        else:
            if config.split == SPLIT:
                primitive_state_left, primitive_state_right = (
                    _reconstruct_at_interface_split(
                        primitive_state,
                        dt,
                        gamma,
                        config,
                        helper_data,
                        registered_variables,
                        axis,
                    )
                )
            else:
                # TODO: improve efficiency
                # this is currently suboptimal, the reconstruction is done for all axes
                # but we only need it for the current axis
                primitives_left_interface, primitives_right_interface = (
                    _reconstruct_at_interface_unsplit(
                        primitive_state,
                        dt,
                        gamma,
                        config,
                        params,
                        helper_data,
                        registered_variables,
                    )
                )
                primitive_state_left = primitives_left_interface[axis - 1]
                primitive_state_right = primitives_right_interface[axis - 1]

        if (
            config.self_gravity_version == RIEMANN_SPLIT
            or config.self_gravity_version == RIEMANN_SPLIT_UNSTABLE
        ):
            # improve code reuse, instead of this copied
            # Riemann solver

            if not (config.riemann_solver == HLLC or config.riemann_solver == HLLC_LM):
                raise NotImplementedError(
                    "The RIEMANN_SPLIT gravity scheme is currently only implemented for HLLC and HLLC_LM."
                )

            rho_L = primitive_state_left[registered_variables.density_index]
            u_L = primitive_state_left[axis]

            rho_R = primitive_state_right[registered_variables.density_index]
            u_R = primitive_state_right[axis]

            p_L = primitive_state_left[registered_variables.pressure_index]
            p_R = primitive_state_right[registered_variables.pressure_index]

            # calculate the sound speeds
            c_L = speed_of_sound(rho_L, p_L, gamma)
            c_R = speed_of_sound(rho_R, p_R, gamma)

            # get the left and right states and fluxes
            F_L = _euler_flux(
                primitive_state_left, gamma, config, registered_variables, axis
            )
            F_R = _euler_flux(
                primitive_state_right, gamma, config, registered_variables, axis
            )

            # Roe average of the velocity
            u_hat = (jnp.sqrt(rho_L) * u_L + jnp.sqrt(rho_R) * u_R) / (
                jnp.sqrt(rho_L) + jnp.sqrt(rho_R)
            )

            # Roe average of the sound speed
            c_hat_squared = (c_L**2 * jnp.sqrt(rho_L) + c_R**2 * jnp.sqrt(rho_R)) / (
                jnp.sqrt(rho_L) + jnp.sqrt(rho_R)
            ) + 0.5 * (
                jnp.sqrt(rho_L)
                * jnp.sqrt(rho_R)
                / (jnp.sqrt(rho_L) + jnp.sqrt(rho_R)) ** 2
            ) * (u_R - u_L) ** 2
            c_hat = jnp.sqrt(c_hat_squared)

            # Einfeldt estimates of maximum left and right signal speeds
            S_L = jnp.minimum(u_L - c_L, u_hat - c_hat)
            S_R = jnp.maximum(u_R + c_R, u_hat + c_hat)

            # contact wave signal speed
            S_star = (
                p_R - p_L + rho_L * u_L * (S_L - u_L) - rho_R * u_R * (S_R - u_R)
            ) / (rho_L * (S_L - u_L) - rho_R * (S_R - u_R))

            # intermediate states
            U_L = conserved_state_from_primitive(
                primitive_state_left, gamma, config, registered_variables
            )
            U_R = conserved_state_from_primitive(
                primitive_state_right, gamma, config, registered_variables
            )

            U_star_L = U_L.at[axis].set(rho_L * S_star)
            U_star_L = U_star_L.at[registered_variables.pressure_index].add(
                (S_star - u_L) * (rho_L * S_star + p_L / (S_L - u_L))
            )
            U_star_L = U_star_L * (S_L - u_L) / (S_L - S_star)

            U_star_R = U_R.at[axis].set(rho_R * S_star)
            U_star_R = U_star_R.at[registered_variables.pressure_index].add(
                (S_star - u_R) * (rho_R * S_star + p_R / (S_R - u_R))
            )
            U_star_R = U_star_R * (S_R - u_R) / (S_R - S_star)

            # HLLC-LM adaptation
            # following
            # https://doi.org/10.1016/j.jcp.2020.109762
            if config.riemann_solver == HLLC_LM:
                Ma_limit = 0.1
                Ma_local = jnp.maximum(jnp.abs(u_L / c_L), jnp.abs(u_R / c_R))
                phi = jnp.sin(jnp.minimum(1, Ma_local / Ma_limit) * jnp.pi / 2)
                S_Llm = S_L * phi
                S_Rlm = S_R * phi

            if config.riemann_solver == HLLC_LM:
                S_Lstar = S_Llm
                S_Rstar = S_Rlm
            else:
                S_Lstar = S_L
                S_Rstar = S_R

            bulk_flux_star = 0.5 * (F_L + F_R)
            dissipation_term_star = 0.5 * (
                S_Lstar * (U_star_L - U_L)
                + jnp.abs(S_star) * (U_star_L - U_star_R)
                + S_Rstar * (U_star_R - U_R)
            )

            F_star = bulk_flux_star + dissipation_term_star

            fluxes = jnp.where(S_L >= 0, F_L, F_star)
            fluxes = jnp.where(S_R <= 0, F_R, fluxes)

            # what cell i accounts for regarding the flux between i-1 and i
            fluxes_i_to_im1 = 0.5 * F_R + jnp.minimum(dissipation_term_star, 0)

            # what cell i-1 accounts for regarding the flux between i-1 and i
            fluxes_im1 = 0.5 * F_L + jnp.maximum(dissipation_term_star, 0)

            if config.self_gravity_version == RIEMANN_SPLIT_UNSTABLE:
                # stable but big spread in specific entropy
                # fluxes_i_to_im1 = jnp.where(S_R <= 0, F_R, fluxes_i_to_im1)
                # fluxes_i_to_im1 = jnp.where(S_L >= 0, 0, fluxes_i_to_im1)
                # fluxes_im1 = jnp.where(S_L >= 0, F_L, fluxes_im1)
                # fluxes_im1 = jnp.where(S_R <= 0, 0, fluxes_im1)

                # less stable but reduced spread

                fluxes_i_to_im1 = jnp.where(S_R <= 0, F_R / 2, fluxes_i_to_im1)
                fluxes_i_to_im1 = jnp.where(S_L >= 0, F_L / 2, fluxes_i_to_im1)
                fluxes_im1 = jnp.where(S_L >= 0, F_L / 2, fluxes_im1)
                fluxes_im1 = jnp.where(S_R <= 0, F_R / 2, fluxes_im1)

            fluxes_i_to_ip1 = jnp.roll(fluxes_im1, shift=-1, axis=axis)

        elif config.self_gravity_version == DONOR_ACCOUNTING:
            # at index i, the fluxes array contains the flux from i-1 to i
            fluxes = _riemann_solver(
                primitive_state_left,
                primitive_state_right,
                primitive_state,
                gamma,
                config,
                registered_variables,
                axis,
            )
            fluxes_i_to_ip1 = jnp.maximum(jnp.roll(fluxes, shift=-1, axis=axis), 0)
            fluxes_i_to_im1 = jnp.minimum(fluxes, 0)

        elif config.self_gravity_version == HALF_SPLIT:
            # at index i, the fluxes array contains the flux from i-1 to i
            fluxes = _riemann_solver(
                primitive_state_left,
                primitive_state_right,
                primitive_state,
                gamma,
                config,
                registered_variables,
                axis,
            )
            fluxes_i_to_ip1 = 0.5 * jnp.roll(fluxes, shift=-1, axis=axis)
            fluxes_i_to_im1 = 0.5 * fluxes

        acc_backward = (
            -_stencil_add(
                gravitational_potential,
                indices=(0, -1),
                factors=(1.0, -1.0),
                axis=axis - 1,
            )
            / grid_spacing
        )
        acc_forward = (
            -_stencil_add(
                gravitational_potential,
                indices=(1, 0),
                factors=(1.0, -1.0),
                axis=axis - 1,
            )
            / grid_spacing
        )

        fluxes_acc = fluxes_i_to_im1 * acc_backward + fluxes_i_to_ip1 * acc_forward

        source_term = source_term.at[registered_variables.pressure_index].set(
            fluxes_acc[0]
        )

    return source_term


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _apply_self_gravity(
    primitive_state: STATE_TYPE,
    old_primitive_state: STATE_TYPE,
    config: SimulationConfig,
    params: SimulationParams,
    registered_variables: RegisteredVariables,
    helper_data: HelperData,
    gamma: Union[float, Float[Array, ""]],
    gravitational_constant: Union[float, Float[Array, ""]],
    dt: Union[float, Float[Array, ""]],
) -> STATE_TYPE:
    rho = old_primitive_state[registered_variables.density_index]

    potential = _compute_gravitational_potential(
        rho, config.grid_spacing, config, gravitational_constant
    )

    source_term = jnp.zeros_like(primitive_state)

    for i in range(config.dimensionality):
        source_term = source_term + _gravitational_source_term_along_axis(
            potential,
            old_primitive_state,
            config.grid_spacing,
            registered_variables,
            dt,
            gamma,
            config,
            params,
            helper_data,
            i + 1,
        )

    conserved_state = conserved_state_from_primitive(
        primitive_state, gamma, config, registered_variables
    )

    conserved_state = conserved_state + dt * source_term

    primitive_state = primitive_state_from_conserved(
        conserved_state, gamma, config, registered_variables
    )


    if config.boundary_handling == GHOST_CELLS:
        primitive_state = _boundary_handler(primitive_state, config)

    return primitive_state


from typing import NamedTuple
import jax.numpy as jnp
from functools import partial
import jax

from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker

from astronomix.option_classes.simulation_config import (
    GAS_STATE,
    MAGNETIC_FIELD_ONLY,
    MHD_JET_BOUNDARY,
    OPEN_BOUNDARY,
    PERIODIC_BOUNDARY,
    REFLECTIVE_BOUNDARY,
    STATE_TYPE,
    VELOCITY_ONLY,
    SimulationConfig,
)

# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["axis", "set_index", "get_index"])
def _set_along_axis(
    primitive_state: STATE_TYPE, axis: int, set_index: int, get_index: int
) -> STATE_TYPE:
    s_set = (
        (slice(None),) * axis
        + (set_index,)
        + (slice(None),) * (primitive_state.ndim - axis - 1)
    )
    s_get = (
        (slice(None),) * axis
        + (get_index,)
        + (slice(None),) * (primitive_state.ndim - axis - 1)
    )

    primitive_state = primitive_state.at[s_set].set(primitive_state[s_get])

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit, static_argnames=["axis", "set_index", "get_index", "var_index", "factor"]
)
def _set_specific_var_along_axis(
    primitive_state: STATE_TYPE,
    axis: int,
    set_index: int,
    get_index: int,
    var_index: int,
    factor: float,
) -> STATE_TYPE:
    s_set = (
        (var_index,)
        + (slice(None),) * (axis - 1)
        + (set_index,)
        + (slice(None),) * (primitive_state.ndim - axis - 1)
    )
    s_get = (
        (var_index,)
        + (slice(None),) * (axis - 1)
        + (get_index,)
        + (slice(None),) * (primitive_state.ndim - axis - 1)
    )

    primitive_state = primitive_state.at[s_set].set(factor * primitive_state[s_get])

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "type_handled"])
def _boundary_handler(
    primitive_state: STATE_TYPE,
    config: SimulationConfig,
    type_handled: int = GAS_STATE,
) -> STATE_TYPE:
    """Apply the boundary conditions to the primitive states.

    Args:
        primitive_state: The primitive state array.
        left_boundary: The left boundary condition.
        right_boundary: The right boundary condition.

    Returns:
        The primitive state array with the boundary conditions applied.
    """

    if config.dimensionality == 1:
        if config.boundary_settings.left_boundary == OPEN_BOUNDARY:
            primitive_state = _open_left_boundary(
                primitive_state, config.num_ghost_cells, axis=1
            )
        elif config.boundary_settings.left_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_left_boundary1d(primitive_state)

        if config.boundary_settings.right_boundary == OPEN_BOUNDARY:
            primitive_state = _open_right_boundary(
                primitive_state, config.num_ghost_cells, axis=1
            )
        elif config.boundary_settings.right_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_right_boundary1d(primitive_state)

        if (
            config.boundary_settings.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.right_boundary == PERIODIC_BOUNDARY
        ):
            primitive_state = _periodic_boundaries(
                primitive_state, config.num_ghost_cells, axis=1
            )

    if config.dimensionality == 2:
        if config.boundary_settings.x.left_boundary == OPEN_BOUNDARY:
            primitive_state = _open_left_boundary(
                primitive_state, config.num_ghost_cells, axis=1
            )

        if config.boundary_settings.x.right_boundary == OPEN_BOUNDARY:
            primitive_state = _open_right_boundary(
                primitive_state, config.num_ghost_cells, axis=1
            )

        if config.boundary_settings.y.left_boundary == OPEN_BOUNDARY:
            primitive_state = _open_left_boundary(
                primitive_state, config.num_ghost_cells, axis=2
            )

        if config.boundary_settings.y.right_boundary == OPEN_BOUNDARY:
            primitive_state = _open_right_boundary(
                primitive_state, config.num_ghost_cells, axis=2
            )

        if config.boundary_settings.x.left_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_left_boundary(
                primitive_state, config.num_ghost_cells, axis=1
            )

        if config.boundary_settings.x.right_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_right_boundary(
                primitive_state, config.num_ghost_cells, axis=1
            )

        if config.boundary_settings.y.left_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_left_boundary(
                primitive_state, config.num_ghost_cells, axis=2
            )

        if config.boundary_settings.y.right_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_right_boundary(
                primitive_state, config.num_ghost_cells, axis=2
            )

        if (
            config.boundary_settings.x.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.x.right_boundary == PERIODIC_BOUNDARY
        ):
            primitive_state = _periodic_boundaries(
                primitive_state, config.num_ghost_cells, axis=1
            )

        if (
            config.boundary_settings.y.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.y.right_boundary == PERIODIC_BOUNDARY
        ):
            primitive_state = _periodic_boundaries(
                primitive_state, config.num_ghost_cells, axis=2
            )

        if config.boundary_settings.y.left_boundary == MHD_JET_BOUNDARY:
            primitive_state = _jet_left_boundary(
                primitive_state,
                config.num_ghost_cells,
                axis=2,
                grid_spacing=config.grid_spacing,
                num_cells=config.num_cells,
                type_handled=type_handled,
            )

    if config.dimensionality == 3:
        if config.boundary_settings.x.left_boundary == OPEN_BOUNDARY:
            primitive_state = _open_left_boundary(
                primitive_state, config.num_ghost_cells, axis=1
            )

        if config.boundary_settings.x.right_boundary == OPEN_BOUNDARY:
            primitive_state = _open_right_boundary(
                primitive_state, config.num_ghost_cells, axis=1
            )

        if config.boundary_settings.y.left_boundary == OPEN_BOUNDARY:
            primitive_state = _open_left_boundary(
                primitive_state, config.num_ghost_cells, axis=2
            )

        if config.boundary_settings.y.right_boundary == OPEN_BOUNDARY:
            primitive_state = _open_right_boundary(
                primitive_state, config.num_ghost_cells, axis=2
            )

        if config.boundary_settings.z.left_boundary == OPEN_BOUNDARY:
            primitive_state = _open_left_boundary(
                primitive_state, config.num_ghost_cells, axis=3
            )

        if config.boundary_settings.z.right_boundary == OPEN_BOUNDARY:
            primitive_state = _open_right_boundary(
                primitive_state, config.num_ghost_cells, axis=3
            )

        if config.boundary_settings.x.left_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_left_boundary(
                primitive_state, config.num_ghost_cells, axis=1
            )

        if config.boundary_settings.x.right_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_right_boundary(
                primitive_state, config.num_ghost_cells, axis=1
            )

        if config.boundary_settings.y.left_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_left_boundary(
                primitive_state, config.num_ghost_cells, axis=2
            )

        if config.boundary_settings.y.right_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_right_boundary(
                primitive_state, config.num_ghost_cells, axis=2
            )

        if config.boundary_settings.z.left_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_left_boundary(
                primitive_state, config.num_ghost_cells, axis=3
            )

        if config.boundary_settings.z.right_boundary == REFLECTIVE_BOUNDARY:
            primitive_state = _reflective_right_boundary(
                primitive_state, config.num_ghost_cells, axis=3
            )

        if (
            config.boundary_settings.x.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.x.right_boundary == PERIODIC_BOUNDARY
        ):
            primitive_state = _periodic_boundaries(
                primitive_state, config.num_ghost_cells, axis=1
            )

        if (
            config.boundary_settings.y.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.y.right_boundary == PERIODIC_BOUNDARY
        ):
            primitive_state = _periodic_boundaries(
                primitive_state, config.num_ghost_cells, axis=2
            )

        if (
            config.boundary_settings.z.left_boundary == PERIODIC_BOUNDARY
            and config.boundary_settings.z.right_boundary == PERIODIC_BOUNDARY
        ):
            primitive_state = _periodic_boundaries(
                primitive_state, config.num_ghost_cells, axis=3
            )

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["axis", "num_ghost_cells"])
def _open_right_boundary(
    primitive_state: STATE_TYPE, num_ghost_cells: int, axis: int
) -> STATE_TYPE:

    get_index = -num_ghost_cells - 1

    for set_index in range(-num_ghost_cells, 0):
        primitive_state = _set_along_axis(primitive_state, axis, set_index, get_index)

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["axis", "num_ghost_cells"])
def _open_left_boundary(
    primitive_state: STATE_TYPE, num_ghost_cells: int, axis: int
) -> STATE_TYPE:

    get_index = num_ghost_cells

    for set_index in range(num_ghost_cells):
        primitive_state = _set_along_axis(primitive_state, axis, set_index, get_index)

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["axis", "num_ghost_cells"])
def _periodic_boundaries(
    primitive_state: STATE_TYPE, num_ghost_cells: int, axis: int
) -> STATE_TYPE:

    # primitive_state = _set_along_axis(primitive_state, axis, 0, -4)
    # primitive_state = _set_along_axis(primitive_state, axis, 1, -3)

    # primitive_state = _set_along_axis(primitive_state, axis, -2, 2)
    # primitive_state = _set_along_axis(primitive_state, axis, -1, 3)

    for i in range(num_ghost_cells):
        # left boundary
        set_index = i
        get_index = i - 2 * num_ghost_cells
        primitive_state = _set_along_axis(primitive_state, axis, set_index, get_index)

        # right boundary
        set_index = -i - 1
        get_index = 2 * num_ghost_cells - i - 1
        primitive_state = _set_along_axis(primitive_state, axis, set_index, get_index)

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["axis", "num_ghost_cells"])
def _reflective_left_boundary(
    primitive_state: STATE_TYPE, num_ghost_cells: int, axis: int
) -> STATE_TYPE:
    # reflect the velocity component along the axis
    for i in range(num_ghost_cells):
        get_index = i + num_ghost_cells
        set_index = num_ghost_cells - i - 1

        primitive_state = _set_along_axis(primitive_state, axis, set_index, get_index)
        primitive_state = _set_specific_var_along_axis(
            primitive_state, axis, set_index, get_index, axis, -1.0
        )

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["axis", "num_ghost_cells"])
def _reflective_right_boundary(
    primitive_state: STATE_TYPE, num_ghost_cells: int, axis: int
) -> STATE_TYPE:
    # reflect the velocity component along the axis
    for i in range(num_ghost_cells):
        get_index = -i - num_ghost_cells - 1
        set_index = -num_ghost_cells + i

        primitive_state = _set_along_axis(primitive_state, axis, set_index, get_index)
        primitive_state = _set_specific_var_along_axis(
            primitive_state, axis, set_index, get_index, axis, -1.0
        )

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@jax.jit
def _reflective_left_boundary1d(
    primitive_state: Float[Array, "num_vars num_cells"],
) -> Float[Array, "num_vars num_cells"]:

    primitive_state = primitive_state.at[:, 0].set(primitive_state[:, 2])
    primitive_state = primitive_state.at[1, 0].set(-primitive_state[1, 2])

    primitive_state = primitive_state.at[:, 1].set(primitive_state[:, 2])
    primitive_state = primitive_state.at[1, 1].set(-primitive_state[1, 2])

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@jax.jit
def _reflective_right_boundary1d(
    primitive_state: Float[Array, "num_vars num_cells"],
) -> Float[Array, "num_vars num_cells"]:

    primitive_state = primitive_state.at[:, -1].set(primitive_state[:, -3])
    primitive_state = primitive_state.at[1, -1].set(-primitive_state[1, -3])

    primitive_state = primitive_state.at[:, -2].set(primitive_state[:, -3])
    primitive_state = primitive_state.at[1, -2].set(-primitive_state[1, -3])

    return primitive_state


# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit,
    static_argnames=[
        "axis",
        "num_ghost_cells",
        "grid_spacing",
        "num_cells",
        "type_handled",
    ],
)
def _jet_left_boundary(
    primitive_state: STATE_TYPE,
    num_ghost_cells: int,
    axis: int,
    grid_spacing: float,
    num_cells: int,
    type_handled: int,
) -> STATE_TYPE:
    get_index = num_ghost_cells

    for set_index in range(num_ghost_cells):
        primitive_state = _set_along_axis(primitive_state, axis, set_index, get_index)

    half_inj_width = 0.025
    half_inj_cell_num = int(half_inj_width / grid_spacing)

    B0 = 200**0.5
    # to_set = jnp.array([5/3, 0.0, 800.0, 0.0, 1.0, 0.0, B0, 0.0])
    #                     rho, v_x, v_y,   v_z, p,   B_x, B_y, B_z
    to_set_gas_state = jnp.array([5 / 3, 800.0, 0.0, 1.0])
    to_set_velocity = jnp.array([800.0, 0.0, 0.0])
    to_set_magnetic_field = jnp.array([B0, 0.0, 0.0])

    if type_handled == GAS_STATE:
        primitive_state = primitive_state.at[
            :,
            0:num_ghost_cells,
            (num_cells // 2 - half_inj_cell_num) : (num_cells // 2 + half_inj_cell_num),
        ].set(to_set_gas_state[:, None, None])
    elif type_handled == VELOCITY_ONLY:
        primitive_state = primitive_state.at[
            :,
            0:num_ghost_cells,
            (num_cells // 2 - half_inj_cell_num) : (num_cells // 2 + half_inj_cell_num),
        ].set(to_set_velocity[:, None, None])
    elif type_handled == MAGNETIC_FIELD_ONLY:
        primitive_state = primitive_state.at[
            :,
            0:num_ghost_cells,
            (num_cells // 2 - half_inj_cell_num) : (num_cells // 2 + half_inj_cell_num),
        ].set(to_set_magnetic_field[:, None, None])

    return primitive_state



# general
from typing import Union
import jax
import jax.numpy as jnp
from functools import partial

# typing
from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker
from typing import Union
from astronomix.option_classes.simulation_config import STATE_TYPE

# astronomix classes
from astronomix._physics_modules._cosmic_rays.cosmic_ray_options import CosmicRayParams
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import SimulationConfig

# astronomix constants
from astronomix.option_classes.simulation_config import SPHERICAL

# astronomix functions
from astronomix.shock_finder.shock_finder import find_shock_zone

# NOTE: currently only supports 1d setups, TODO: generalize


@partial(jax.jit, static_argnames=["registered_variables", "config"])
def inject_crs_at_strongest_shock(
    primitive_state: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    helper_data: HelperData,
    cosmic_ray_params: CosmicRayParams,
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    dt: Union[float, Float[Array, ""]],
) -> STATE_TYPE:
    """
    Cosmic ray injection at shock fronts.
    Currently only at the strongest shock in the domain.

    The implementation generally follows

    Pfrommer, Christoph, et al. "Simulating cosmic ray physics on a moving mesh."
    Monthly Notices of the Royal Astronomical Society 465.4 (2017): 4500-4529.
    https://arxiv.org/abs/1604.07399

    and

    Dubois, Yohan, et al. "Shock-accelerated cosmic rays and streaming instability
    in the adaptive mesh refinement code Ramses."
    Astronomy & Astrophysics 631 (2019): A121.
    https://arxiv.org/abs/1907.04300

    Args:
        primitive_state: The primitive state array.
        gamma: The adiabatic index.
        helper_data: The helper data.
        cosmic_ray_params: The cosmic ray parameters.
        config: The simulation configuration.
        registered_variables: The registered variables.
        dt: The time step.

    Returns:
        The primitive state array with injected cosmic rays.

    """

    num_cells = primitive_state.shape[1]

    # the injection efficiency is specified by the user
    injection_efficiency = cosmic_ray_params.diffusive_shock_acceleration_efficiency
    # in future use e.g. models as implemented in
    # https://github.com/LudwigBoess/DiffusiveShockAccelerationModels.jl/tree/main/src/mach_models

    # currently for crs the adiabatic indices are hard coded
    gamma_cr = 4 / 3
    gamma_gas = gamma

    # find the strongest shock
    max_shock_idx, left_idx, right_idx = find_shock_zone(
        primitive_state, config, registered_variables, helper_data
    )

    # left_idx = left_idx + 2
    # +2 leads to a smoother transition of the different pressure
    # components in the shock, but will lead to problems at lower
    # resolutions, also note the problem that CR pressure injected
    # in the broadened shock layer experiences PCR∇.u forces
    # (Dubois et al, 2019), so one might overinject effectively

    # we only consider a shock moving from left to right
    # pre-shock is upstream in the shock frame
    # post-shock is downstream in the shock frame
    pre_shock_idx = right_idx + 1
    post_shock_idx = left_idx - 1

    # get the pre shock (upstream) state
    rho1 = primitive_state[registered_variables.density_index, pre_shock_idx]  # density
    P1 = primitive_state[
        registered_variables.pressure_index, pre_shock_idx
    ]  # total pressure
    P1_CRs = (
        primitive_state[registered_variables.cosmic_ray_n_index, pre_shock_idx]
        ** gamma_cr
    )  # cosmic ray pressure
    P1_gas = P1 - P1_CRs  # gas pressure
    e1_gas = P1_gas / (gamma_gas - 1)  # gas energy / volume
    e1_crs = P1_CRs / (gamma_cr - 1)  # cosmic ray energy / volume
    e1 = e1_gas + e1_crs  # total energy / volume

    # get the post shock state
    rho2 = primitive_state[registered_variables.density_index, post_shock_idx]
    P2 = primitive_state[registered_variables.pressure_index, post_shock_idx]
    P2_CRs = (
        primitive_state[registered_variables.cosmic_ray_n_index, post_shock_idx]
        ** gamma_cr
    )
    P2_gas = P2 - P2_CRs
    e2_gas = P2_gas / (gamma_gas - 1)
    e2_crs = P2_CRs / (gamma_cr - 1)
    e2 = e2_gas + e2_crs

    # get the effective adiabatic index
    gamma_eff1 = (gamma_cr * P1_CRs + gamma_gas * P1_gas) / P1

    # calculate the pre-shock sound speed
    c1 = jnp.sqrt(gamma_eff1 * P1 / rho1)

    # density ratio
    x_s = rho2 / rho1

    # pre-shock mach number, simplest formula
    # M_1_sq = (P2 / P1 - 1) * x_s / (gamma_eff1 * (x_s - 1))

    gamma_eff1 = (gamma_cr * P1_CRs + gamma_gas * P1_gas) / P1
    gamma_eff2 = (gamma_cr * P2_CRs + gamma_gas * P2_gas) / P2

    gamma1 = P1 / e1 + 1
    gamma2 = P2 / e2 + 1

    gammat = P2 / P1

    C = ((gamma2 + 1) * gammat + gamma2 - 1) * (gamma1 - 1)

    # formula 16 in Dubois et al 2019, differing slightly
    # from the expression in Pfrommer et al 2017, note however
    # that in Pfrommer et al 2017 this formula is only used
    # for the shock finder (because it is only a lower bound)
    # and not the injection itself, where
    # M_1_sq = (P2 / P1 - 1) * x_s / (gamma_eff1 * (x_s - 1))
    # is used, which in my experience led to more crashes
    # in spherical geometry setups
    M_1_sq = (
        1
        / gamma_eff2
        * (gammat - 1)
        * C
        / (C - ((gamma1 + 1) + (gamma1 - 1) * gammat) * (gamma2 - 1))
    )

    # dissipated energy density
    e_diss = e2_gas - e1_gas * x_s**gamma_gas + e2_crs - e1_crs * x_s**gamma_cr

    # dissipated flux
    f_diss = e_diss * jnp.sqrt(M_1_sq) * c1 / x_s

    # calculate the shock surface
    if config.geometry == SPHERICAL:
        shock_radius = helper_data.geometric_centers[max_shock_idx]
        shock_surface = 4 * jnp.pi * shock_radius**2
    else:
        shock_surface = config.grid_spacing ** (config.dimensionality - 1)

    # energy to be injected in the form of cosmic ray pressure
    DeltaE_CR = f_diss * shock_surface * dt * injection_efficiency

    # get a mask for the shock, if an injection as done in Pfommer et al. 2017 is desired
    indices = jnp.arange(num_cells)

    # in Pfrommer et al 2017 instead of left_idx the post_shock_idx is used
    # as far as I understand
    shock_zone_mask = (indices >= left_idx) & (indices <= max_shock_idx)

    cosmic_ray_pressure = (
        primitive_state[registered_variables.cosmic_ray_n_index] ** gamma_cr
    )
    gas_pressure = (
        primitive_state[registered_variables.pressure_index] - cosmic_ray_pressure
    )
    e_th = gas_pressure / (gamma_gas - 1)
    e_cr = cosmic_ray_pressure / (gamma_cr - 1)
    E_tot = (e_th + e_cr) * helper_data.cell_volumes
    DeltaEtot = jnp.sum(jnp.where(shock_zone_mask, E_tot - E_tot[right_idx], 0))
    DeltaE_CR_split = DeltaE_CR * (E_tot - E_tot[right_idx]) / DeltaEtot
    DeltaE_CR_split = jnp.where(shock_zone_mask, DeltaE_CR_split, 0)

    # to be injected cosmic ray pressure
    p_cr_injection = (
        primitive_state[registered_variables.cosmic_ray_n_index] ** gamma_cr
    )
    # updated cosmic ray pressure
    p_cr_injection_new = p_cr_injection + DeltaE_CR_split / helper_data.cell_volumes * (
        gamma_cr - 1
    )
    # note that we work with n_cr = P_CR ^ (1 / gamma_cr) to describe the cosmic rays
    n_cr_injection_new = p_cr_injection_new ** (1 / gamma_cr)
    primitive_state = primitive_state.at[registered_variables.cosmic_ray_n_index].set(
        n_cr_injection_new
    )

    # we want energy and not pressure conservation, so the total pressure must be adapted
    delta_p_gas = DeltaE_CR_split / helper_data.cell_volumes * (gamma_gas - 1)
    p_gas_new = (
        primitive_state[registered_variables.pressure_index]
        - p_cr_injection
        - delta_p_gas
    )
    total_pressure_new = p_gas_new + p_cr_injection_new

    # update the total pressure
    primitive_state = primitive_state.at[registered_variables.pressure_index].set(
        total_pressure_new
    )

    return primitive_state



import jax.numpy as jnp
import jax
from functools import partial

from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker

from typing import Union

from astronomix._physics_modules._cosmic_rays.cr_fluid_equations import (
    total_energy_from_primitives_with_crs,
    total_pressure_from_conserved_with_crs,
)
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import (
    FIELD_TYPE,
    STATE_TYPE,
    SimulationConfig,
)
from astronomix.option_classes.simulation_params import SimulationParams

# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def primitive_state_from_conserved(
    conserved_state: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    """Convert the conserved state to the primitive state.

    Args:
        conserved_state: The conserved state.
        gamma: The adiabatic index of the fluid.

    Returns:
        The primitive state.
    """
    # note the indices of the conserved variables
    # are the same as the indices of the primitive variables
    # so velocity and moentum density have the same index

    rho = conserved_state[registered_variables.density_index]
    E = conserved_state[registered_variables.pressure_index]

    if config.dimensionality == 1:
        u = conserved_state[registered_variables.velocity_index] / rho
    elif config.dimensionality == 2:
        ux = conserved_state[registered_variables.velocity_index.x] / rho
        uy = conserved_state[registered_variables.velocity_index.y] / rho
        # TODO: find smarter way, problem d/dx sqrt(0) = inf
        u = jnp.sqrt(ux**2 + uy**2 + 1e-20)
    elif config.dimensionality == 3:
        ux = conserved_state[registered_variables.velocity_index.x] / rho
        uy = conserved_state[registered_variables.velocity_index.y] / rho
        uz = conserved_state[registered_variables.velocity_index.z] / rho
        u = jnp.sqrt(ux**2 + uy**2 + uz**2 + 1e-20)

    p = pressure_from_energy(E, rho, u, gamma)

    if registered_variables.cosmic_ray_n_active:
        p = total_pressure_from_conserved_with_crs(
            conserved_state, registered_variables
        )
    else:
        p = pressure_from_energy(E, rho, u, gamma)

    # set the primitive state
    primitive_state = conserved_state.at[registered_variables.pressure_index].set(p)

    if config.dimensionality == 1:
        primitive_state = primitive_state.at[registered_variables.velocity_index].set(u)
    elif config.dimensionality == 2:
        primitive_state = primitive_state.at[registered_variables.velocity_index.x].set(
            ux
        )
        primitive_state = primitive_state.at[registered_variables.velocity_index.y].set(
            uy
        )
    elif config.dimensionality == 3:
        primitive_state = primitive_state.at[registered_variables.velocity_index.x].set(
            ux
        )
        primitive_state = primitive_state.at[registered_variables.velocity_index.y].set(
            uy
        )
        primitive_state = primitive_state.at[registered_variables.velocity_index.z].set(
            uz
        )

    # for all other variables assume that primitive and conserved state are the same
    # as for the mass density

    return primitive_state


# ===========================================

# ======= Create the conserved state ========


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def conserved_state_from_primitive(
    primitive_state: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    """Convert the primitive state to the conserved state.

    Args:
        primitive_state: The primitive state.
        gamma: The adiabatic index of the fluid.

    Returns:
        The conserved state.
    """

    rho = primitive_state[registered_variables.density_index]

    u = get_absolute_velocity(primitive_state, config, registered_variables)
    p = primitive_state[registered_variables.pressure_index]

    if registered_variables.cosmic_ray_n_active:
        E = total_energy_from_primitives_with_crs(primitive_state, registered_variables)
    else:
        E = total_energy_from_primitives(rho, u, p, gamma)

    conserved_state = primitive_state.at[registered_variables.pressure_index].set(E)

    if config.dimensionality == 1:
        conserved_state = conserved_state.at[registered_variables.velocity_index].set(
            rho * primitive_state[registered_variables.velocity_index]
        )
    elif config.dimensionality == 2:
        conserved_state = conserved_state.at[registered_variables.velocity_index.x].set(
            rho * primitive_state[registered_variables.velocity_index.x]
        )
        conserved_state = conserved_state.at[registered_variables.velocity_index.y].set(
            rho * primitive_state[registered_variables.velocity_index.y]
        )
    elif config.dimensionality == 3:
        conserved_state = conserved_state.at[registered_variables.velocity_index.x].set(
            rho * primitive_state[registered_variables.velocity_index.x]
        )
        conserved_state = conserved_state.at[registered_variables.velocity_index.y].set(
            rho * primitive_state[registered_variables.velocity_index.y]
        )
        conserved_state = conserved_state.at[registered_variables.velocity_index.z].set(
            rho * primitive_state[registered_variables.velocity_index.z]
        )
    else:
        raise ValueError("Invalid dimension.")

    return conserved_state

# ===========================================

# =============== Fluid physics ===============


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=["config", "registered_variables"])
def get_absolute_velocity(
    primitive_state: STATE_TYPE,
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
) -> Union[
    Float[Array, "num_cells"],
    Float[Array, "num_cells_x num_cells_y"],
    Float[Array, "num_cells_x num_cells_y num_cells_z"],
]:
    """Get the absolute velocity of the fluid.

    Args:
        primitive_state: The primitive state of the fluid.
        config: The simulation configuration.
        registered_variables: The registered variables.

    Returns:
        The absolute velocity.
    """
    if config.dimensionality == 1:
        return jnp.abs(primitive_state[registered_variables.velocity_index])
    elif config.dimensionality == 2:
        return jnp.sqrt(
            primitive_state[registered_variables.velocity_index.x] ** 2
            + primitive_state[registered_variables.velocity_index.y] ** 2
        )
    elif config.dimensionality == 3:
        return jnp.sqrt(
            primitive_state[registered_variables.velocity_index.x] ** 2
            + primitive_state[registered_variables.velocity_index.y] ** 2
            + primitive_state[registered_variables.velocity_index.z] ** 2
            + 1e-20
        )
    else:
        raise ValueError("Invalid dimension.")


@jax.jit
def pressure_from_internal_energy(e, rho, gamma):
    """
    Calculate the pressure from the internal energy.

    Args:
        e: The internal energy.
        rho: The density.
        gamma: The adiabatic index.

    Returns:
        The pressure.
    """
    return (gamma - 1) * rho * e


@jax.jit
def internal_energy_from_energy(E, rho, u):
    """Calculate the internal energy from the total energy.

    Args:
        E: The total energy.
        rho: The density.
        u: The velocity.

    Returns:
        The internal energy.
    """
    return E / rho - 0.5 * u**2


@jax.jit
def pressure_from_energy(E, rho, u, gamma):
    """Calculate the pressure from the total energy.

    Args:
        E: The total energy.
        rho: The density.
        u: The velocity.
        gamma: The adiabatic index.

    Returns:
        The pressure.
    """

    e = internal_energy_from_energy(E, rho, u)
    return pressure_from_internal_energy(e, rho, gamma)



@jax.jit
def total_energy_from_primitives(rho, u, p, gamma):
    """Calculate the total energy from the primitive variables.

    Args:
        rho: The density.
        u: The velocity.
        p: The pressure.
        gamma: The adiabatic index.

    Returns:
        The total energy.
    """

    return p / (gamma - 1) + 0.5 * rho * u**2

@jax.jit
def speed_of_sound(rho, p, gamma):
    """Calculate the speed of sound.

    Args:
        rho: The density.
        p: The pressure.
        gamma: The adiabatic index.

    Returns:
        The speed of sound.
    """
    return jnp.sqrt(gamma * p / rho)


# ===========================================



# general
from types import NoneType
import jax
import jax.numpy as jnp
from functools import partial

from equinox.internal._loop.checkpointed import checkpointed_while_loop

# type checking
from jaxtyping import jaxtyped
from beartype import beartype as typechecker
from typing import Union

# runtime debugging
from jax.experimental import checkify

# astronomix constants
from astronomix._finite_difference._maths._differencing import _interface_field_divergence
from astronomix._finite_difference._state_evolution._evolve_state import _evolve_state_fd
from astronomix._finite_difference._timestep_estimation._timestep_estimator import _cfl_time_step_fd
from astronomix._finite_volume._magnetic_update._vector_maths import divergence3D
from astronomix._geometry.boundaries import _boundary_handler
from astronomix._physics_modules._turbulent_forcing._turbulent_forcing import _apply_forcing
from astronomix.data_classes.simulation_state_struct import StateStruct
from astronomix.option_classes.simulation_config import BACKWARDS, FINITE_DIFFERENCE, FINITE_VOLUME, FORWARDS, GHOST_CELLS, STATE_TYPE

# astronomix containers
from astronomix.option_classes.simulation_config import SimulationConfig
from astronomix.data_classes.simulation_helper_data import HelperData, get_helper_data
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_params import SimulationParams
from astronomix.data_classes.simulation_snapshot_data import SnapshotData

# astronomix functions
from astronomix._finite_volume._state_evolution.evolve_state import _evolve_state_fv
from astronomix._physics_modules.run_physics_modules import _run_physics_modules
from astronomix._finite_volume._timestep_estimation._timestep_estimator import (
    _cfl_time_step,
    _source_term_aware_time_step,
)
from astronomix._fluid_equations.total_quantities import (
    calculate_internal_energy,
    calculate_radial_momentum,
    calculate_total_mass,
)
from astronomix._fluid_equations.total_quantities import (
    calculate_total_energy,
    calculate_kinetic_energy,
    calculate_gravitational_energy,
)
from astronomix.time_stepping._utils import _pad, _unpad

# progress bar
from astronomix.time_stepping._progress_bar import _show_progress

# timing
from timeit import default_timer as timer


# @jaxtyped(typechecker=typechecker)
def time_integration(
    primitive_state: STATE_TYPE,
    config: SimulationConfig,
    params: SimulationParams,
    registered_variables: RegisteredVariables,
    snapshot_callable = None,
    sharding: Union[NoneType, jax.NamedSharding] = None,
) -> Union[STATE_TYPE, SnapshotData]:
    """
    Integrate the fluid equations in time. For the options of
    the time integration see the simulation configuration and
    the simulation parameters.

    Args:
        primitive_state: The primitive state array.
        config: The simulation configuration.
        params: The simulation parameters.
        registered_variables: The registered variables.
        snapshot_callable: A callable which is called at certain time points
            if config.activate_snapshot_callback is True. The callable must
            have the signature
                callable(time: float, state: STATE_TYPE, registered_variables: RegisteredVariables) -> None
            and can be used to e.g. output the current state to disk or
            directly produce intermediate plots. Note that inside the callable,
            to pass data to memory, one must use
                jax.debug.callback(
                    function, args...
                )
            To avoid moving large amounts of data to the host, only pass
            the necessary data to the function in the jax.debug.callback call,
            e.g. only the slice or summary statistics you need.
        sharding: The sharding to use for the padded helper data. If None,
                  no sharding is applied.

    Returns:
        Depending on the configuration (return_snapshots, num_snapshots)
        either the final state of the fluid after the time
        integration of snapshots of the time evolution.

    """

    # Here we prepare everything for the actual time integration function,
    # _time_integration, which is jitted below. This includes setting up
    # runtime debugging via checkify if requested, printing the elapsed
    # time if requested, compiling the function for memory analysis if
    # requested, etc.

    helper_data_pad = get_helper_data(
        config,
        sharding,
        padded = config.boundary_handling == GHOST_CELLS,
        production = True
    )
    
    helper_data = get_helper_data(
        config,
        sharding,
        padded = False,
        production = True
    )

    if config.donate_state:
        time_integration_jit = jax.jit(
            _time_integration,
            static_argnames=[
                "config",
                "registered_variables",
                "snapshot_callable"
            ],
            donate_argnames=["state"],
        )
    else:
        time_integration_jit = jax.jit(
            _time_integration,
            static_argnames=[
                "config",
                "registered_variables",
                "snapshot_callable"
            ],
        )

    if config.runtime_debugging:
        errors = (
            checkify.user_checks
            | checkify.index_checks
            | checkify.float_checks
            | checkify.nan_checks
            | checkify.div_checks
        )
        checked_integration = checkify.checkify(_time_integration, errors)

        err, final_state = checked_integration(
            primitive_state,
            config,
            params,
            registered_variables,
            helper_data,
            helper_data_pad,
            snapshot_callable,
        )
        err.throw()

    else:
        if config.memory_analysis:
            compiled_step = time_integration_jit.lower(
                primitive_state,
                config,
                params,
                registered_variables,
                helper_data,
                helper_data_pad,
                snapshot_callable,
            ).compile()
            compiled_stats = compiled_step.memory_analysis()
            if compiled_stats is not None:
                # Calculate total memory usage including temporary storage,
                # arguments, and outputs (but excluding aliases)
                total = (
                    compiled_stats.temp_size_in_bytes
                    + compiled_stats.argument_size_in_bytes
                    + compiled_stats.output_size_in_bytes
                    - compiled_stats.alias_size_in_bytes
                )
                print("=== Compiled memory usage PER DEVICE ===")
                print(
                    f"Temp size: {compiled_stats.temp_size_in_bytes / (1024**2):.2f} MB"
                )
                print(
                    f"Argument size: {compiled_stats.argument_size_in_bytes / (1024**2):.2f} MB"
                )
                print(f"Total size: {total / (1024**2):.2f} MB")
                print("========================================")

        if config.print_elapsed_time:
            if not config.memory_analysis:
                # compile the time integration function
                time_integration_jit.lower(
                    primitive_state,
                    config,
                    params,
                    registered_variables,
                    helper_data,
                    helper_data_pad,
                    snapshot_callable,
                ).compile()

            start_time = timer()
            print("🚀 Starting simulation...")

        final_state = time_integration_jit(
            primitive_state,
            config,
            params,
            registered_variables,
            helper_data,
            helper_data_pad,
            snapshot_callable,
        )

        if config.print_elapsed_time:
            if config.return_snapshots and config.snapshot_settings.return_final_state:
                final_state.final_state.block_until_ready()
            else:
                final_state.block_until_ready()
            end_time = timer()
            print("🏁 Simulation finished!")
            print(f"⏱️ Time elapsed: {end_time - start_time:.2f} seconds")
            if config.return_snapshots:
                num_iterations = final_state.num_iterations
                print(f"🔄 Number of iterations: {num_iterations}")
                # print the time per iteration
                print(
                    f"⏱️ / 🔄 time per iteration: {(end_time - start_time) / num_iterations} seconds"
                )

    return final_state


def _time_integration(
    state: Union[STATE_TYPE, StateStruct],
    config: SimulationConfig,
    params: SimulationParams,
    registered_variables: RegisteredVariables,
    helper_data_unpad: Union[HelperData, NoneType],
    helper_data_pad: Union[HelperData, NoneType],
    snapshot_callable = None,
) -> Union[STATE_TYPE, StateStruct, SnapshotData]:
    """
    Time integration.

    Args:
        primitive_state: The primitive state array.
        config: The simulation configuration.
        params: The simulation parameters.
        helper_data: The helper data.

    Returns:
        Depending on the configuration (return_snapshots, num_snapshots)
        either the final state of the fluid after the time integration
        of snapshots of the time evolution.
    """

    # in simulations, where we also follow e.g. star particles,
    # the state may be a struct containing the primitive state
    # and the star particle data
    if config.state_struct:
        primitive_state = state.primitive_state
    else:
        primitive_state = state

    # we must pad the state with ghost cells
    # pad the primitive state with two ghost cells on each side
    # to account for the periodic boundary conditions
    original_shape = primitive_state.shape

    if config.boundary_handling == GHOST_CELLS:
        primitive_state = _pad(primitive_state, config)

        # important for active boundaries influencing
        # the time step criterion for now only gas state
        if config.mhd:
            primitive_state = primitive_state.at[:-3, ...].set(
                _boundary_handler(primitive_state[:-3, ...], config)
            )
        else:
            primitive_state = _boundary_handler(primitive_state, config)

    # -------------------------------------------------------------
    # =============== ↓ Setup of the snapshot array ↓ =============
    # -------------------------------------------------------------

    # In case the user requests the fluid state (or given
    # statistics) at certain time points (and not only a
    # final state at the end), we have to set up the arrays
    # to store this data.

    if config.return_snapshots:
        time_points = jnp.zeros(config.num_snapshots)

        states = (
            jnp.zeros((config.num_snapshots, *original_shape))
            if config.snapshot_settings.return_states
            else None
        )
        total_mass = (
            jnp.zeros(config.num_snapshots)
            if config.snapshot_settings.return_total_mass
            else None
        )
        total_energy = (
            jnp.zeros(config.num_snapshots)
            if config.snapshot_settings.return_total_energy
            else None
        )
        internal_energy = (
            jnp.zeros(config.num_snapshots)
            if config.snapshot_settings.return_internal_energy
            else None
        )
        kinetic_energy = (
            jnp.zeros(config.num_snapshots)
            if config.snapshot_settings.return_kinetic_energy
            else None
        )
        radial_momentum = (
            jnp.zeros(config.num_snapshots)
            if config.snapshot_settings.return_radial_momentum
            else None
        )

        gravitational_energy = (
            jnp.zeros(config.num_snapshots)
            if config.snapshot_settings.return_gravitational_energy
            and config.self_gravity
            else None
        )

        magnetic_divergence = (
            jnp.zeros(config.num_snapshots)
            if config.snapshot_settings.return_magnetic_divergence
            and config.mhd
            else None
        )

        current_checkpoint = 0

        snapshot_data = SnapshotData(
            time_points=time_points,
            states=states,
            total_mass=total_mass,
            total_energy=total_energy,
            internal_energy=internal_energy,
            kinetic_energy=kinetic_energy,
            gravitational_energy=gravitational_energy,
            current_checkpoint=current_checkpoint,
            radial_momentum=radial_momentum,
            magnetic_divergence=magnetic_divergence,
            final_state=None,
        )

    elif config.activate_snapshot_callback:
        current_checkpoint = 0
        snapshot_data = SnapshotData(
            time_points=None,
            states=None,
            total_mass=None,
            total_energy=None,
            current_checkpoint=current_checkpoint,
        )

    # -------------------------------------------------------------
    # =============== ↑ Setup of the snapshot array ↑ =============
    # -------------------------------------------------------------

    # -------------------------------------------------------------
    # ====================== ↓ Update step ↓ ======================
    # -------------------------------------------------------------

    # This is the actual update step of the data handled by the time
    # integration function. In the simplest case, this might just
    # take in the primitive state and return the updated primitive state
    # after a time step. However, the data which actually needs to be
    # updated may be more complex, e.g. the SnapshotData needs to be
    # updated appropriately if snapshots are requested.

    def update_step(carry):
        # --------------- ↓ Carry unpacking+ ↓ ----------------

        # Depending on the configuration, the carry might either contain
        #   - the time, the primitive state and the snapshot data
        #   - only the time and the primitive state

        # We need to appropriately unpack the carry and in case we
        # have snapshot data, we also directly update it here at
        # the beginning of the time step.

        if config.return_snapshots:
            # When SnapshotData is involved, we need to unpack the carry
            # correctly and update the SnapshotData if we are currently
            # at a point in time where we want to take a snapshot.

            time, key, primitive_state, snapshot_data = carry

            def update_snapshot_data(time, primitive_state, snapshot_data):
                time_points = snapshot_data.time_points.at[
                    snapshot_data.current_checkpoint
                ].set(time)

                if config.boundary_handling == GHOST_CELLS:
                    unpad_primitive_state = _unpad(primitive_state, config)
                else:
                    unpad_primitive_state = primitive_state

                if config.snapshot_settings.return_states:
                    states = snapshot_data.states.at[
                        snapshot_data.current_checkpoint
                    ].set(unpad_primitive_state)
                else:
                    states = None

                if config.snapshot_settings.return_total_mass:
                    total_mass = snapshot_data.total_mass.at[
                        snapshot_data.current_checkpoint
                    ].set(
                        calculate_total_mass(unpad_primitive_state, helper_data_unpad, config)
                    )
                else:
                    total_mass = None

                if config.snapshot_settings.return_total_energy:
                    total_energy = snapshot_data.total_energy.at[
                        snapshot_data.current_checkpoint
                    ].set(
                        calculate_total_energy(
                            unpad_primitive_state,
                            helper_data_unpad,
                            params.gamma,
                            params.gravitational_constant,
                            config,
                            registered_variables,
                        )
                    )
                else:
                    total_energy = None

                if config.snapshot_settings.return_internal_energy:
                    internal_energy = snapshot_data.internal_energy.at[
                        snapshot_data.current_checkpoint
                    ].set(
                        calculate_internal_energy(
                            unpad_primitive_state,
                            helper_data_unpad,
                            params.gamma,
                            config,
                            registered_variables,
                        )
                    )
                else:
                    internal_energy = None

                if config.snapshot_settings.return_kinetic_energy:
                    kinetic_energy = snapshot_data.kinetic_energy.at[
                        snapshot_data.current_checkpoint
                    ].set(
                        calculate_kinetic_energy(
                            unpad_primitive_state,
                            helper_data_unpad,
                            config,
                            registered_variables,
                        )
                    )
                else:
                    kinetic_energy = None

                if config.snapshot_settings.return_radial_momentum:
                    radial_momentum = snapshot_data.radial_momentum.at[
                        snapshot_data.current_checkpoint
                    ].set(
                        calculate_radial_momentum(
                            unpad_primitive_state,
                            helper_data_unpad,
                            config,
                            registered_variables,
                        )
                    )
                else:
                    radial_momentum = None

                if (
                    config.self_gravity
                    and config.snapshot_settings.return_gravitational_energy
                ):
                    gravitational_energy = snapshot_data.gravitational_energy.at[
                        snapshot_data.current_checkpoint
                    ].set(
                        calculate_gravitational_energy(
                            unpad_primitive_state,
                            helper_data_unpad,
                            params.gravitational_constant,
                            config,
                            registered_variables,
                        )
                    )
                else:
                    gravitational_energy = None

                magnetic_divergence = snapshot_data.magnetic_divergence.at[
                    snapshot_data.current_checkpoint
                ].set(
                    jnp.max(jnp.abs(_interface_field_divergence(
                        unpad_primitive_state[registered_variables.interface_magnetic_field_index.x],
                        unpad_primitive_state[registered_variables.interface_magnetic_field_index.y],
                        unpad_primitive_state[registered_variables.interface_magnetic_field_index.z],
                        config.grid_spacing,
                    ))) if config.solver_mode == FINITE_DIFFERENCE else
                    jnp.max(jnp.abs(
                        divergence3D(
                            unpad_primitive_state[registered_variables.magnetic_index.x:registered_variables.magnetic_index.z+1],
                            config.grid_spacing,
                        )
                    ))
                ) if config.snapshot_settings.return_magnetic_divergence and config.mhd else None

                current_checkpoint = snapshot_data.current_checkpoint + 1
                snapshot_data = snapshot_data._replace(
                    time_points=time_points,
                    states=states,
                    current_checkpoint=current_checkpoint,
                    total_mass=total_mass,
                    total_energy=total_energy,
                    internal_energy=internal_energy,
                    kinetic_energy=kinetic_energy,
                    gravitational_energy=gravitational_energy,
                    radial_momentum=radial_momentum,
                    magnetic_divergence=magnetic_divergence,
                )
                return snapshot_data

            def dont_update_snapshot_data(time, primitive_state, snapshot_data):
                return snapshot_data

            if config.use_specific_snapshot_timepoints:
                snapshot_data = jax.lax.cond(
                    jnp.abs(
                        time
                        - params.snapshot_timepoints[snapshot_data.current_checkpoint]
                    )
                    < 1e-12,
                    update_snapshot_data,
                    dont_update_snapshot_data,
                    time,
                    primitive_state,
                    snapshot_data,
                )
            else:
                snapshot_data = jax.lax.cond(
                    time
                    >= snapshot_data.current_checkpoint
                    * params.t_end
                    / config.num_snapshots,
                    update_snapshot_data,
                    dont_update_snapshot_data,
                    time,
                    primitive_state,
                    snapshot_data,
                )

            num_iterations = snapshot_data.num_iterations + 1
            snapshot_data = snapshot_data._replace(num_iterations=num_iterations)

        elif config.activate_snapshot_callback:
            # Here we deal with the case where the user passes
            # a callable which is applied at certain time points
            # - e.g. to output the current state to disk or
            # directly produce intermediate plots.

            time, key, primitive_state, snapshot_data = carry

            def update_snapshot_data(snapshot_data):
                current_checkpoint = snapshot_data.current_checkpoint + 1
                snapshot_data = snapshot_data._replace(
                    current_checkpoint=current_checkpoint
                )

                # call the user-defined snapshot callable
                # NOTE: to pass data to memory, one must use
                # jax.debug.callback(
                #     function, args...
                # )
                # inside the snapshot_callable. To avoid moving
                # large amounts of data to the host, only pass
                # the necessary data to the function in the
                # jax.debug.callback call, e.g. only the slice
                # or summary statistics you need.
                snapshot_callable(time, primitive_state, registered_variables)

                return snapshot_data

            def dont_update_snapshot_data(snapshot_data):
                return snapshot_data

            snapshot_data = jax.lax.cond(
                time
                >= snapshot_data.current_checkpoint
                * params.t_end
                / config.num_snapshots,
                update_snapshot_data,
                dont_update_snapshot_data,
                snapshot_data,
            )

            num_iterations = snapshot_data.num_iterations + 1
            snapshot_data = snapshot_data._replace(num_iterations=num_iterations)
        else:
            # This is the simplest case where we only have
            # the time and the primitive state in the carry.
            # We just unpack them accordingly.
            time, key, primitive_state = carry

        # --------------- ↑ Carry unpacking+ ↑ ----------------

        # ---------------- ↓ time step logic ↓ ----------------

        # This is the heart of the time integration function.
        # Here we determine the time step size and then evolve
        # the state and run the physics modules.

        # determine the time step size
        if not config.fixed_timestep:
            if config.solver_mode == FINITE_VOLUME:
                if config.source_term_aware_timestep:
                    dt = jax.lax.stop_gradient(
                        _source_term_aware_time_step(
                            primitive_state,
                            config,
                            params,
                            helper_data_pad,
                            registered_variables,
                            time,
                        )
                    )
                else:
                    dt = jax.lax.stop_gradient(
                        _cfl_time_step(
                            primitive_state,
                            config.grid_spacing,
                            params.dt_max,
                            params.gamma,
                            config,
                            registered_variables,
                            params.C_cfl,
                        )
                    )
            elif config.solver_mode == FINITE_DIFFERENCE:
                dt = jax.lax.stop_gradient(
                    _cfl_time_step_fd(
                        primitive_state,
                        config.grid_spacing,
                        params.dt_max,
                        params.gamma,
                        config,
                        params,
                        registered_variables,
                        params.C_cfl,
                    )
                )
        else:
            dt = params.t_end / config.num_timesteps

        # make sure we exactly hit the snapshot time points
        if config.use_specific_snapshot_timepoints and config.return_snapshots:
            dt = jnp.minimum(
                dt, params.snapshot_timepoints[snapshot_data.current_checkpoint] - time
            )

        # make sure we exactly hit the end time
        if config.exact_end_time and not config.use_specific_snapshot_timepoints:
            dt = jnp.minimum(dt, params.t_end - time)

        # ---------------- ↑ time step logic ↑ ----------------

        # ----------------- ↓ CENTRAL UPDATE ↓ ----------------

        if config.solver_mode == FINITE_VOLUME:
            # run physics modules
            # for now we mainly consider the stellar wind, a constant source term term,
            # so the source is handled via a simple Euler step but generally
            # a higher order method (in a split fashion) may be used
            primitive_state = _run_physics_modules(
                primitive_state,
                dt,
                config,
                params,
                helper_data_pad,
                registered_variables,
                time + dt,
            )

        # turbulence forcing, TODO: move to physics modules
        # NOTE: THE KEY IS CURRENTLY DIRECTLY IN THE CARRY
        # FOR THE CASE WITHOUT SNAPSHOT DATA AND NOT PRESENT
        # IN THE CARRY OTHERWISE. TODO: IMPROVE THIS.
        if config.turbulent_forcing_config.turbulent_forcing:
            key, primitive_state = _apply_forcing(
                key,
                primitive_state,
                dt,
                params.turbulent_forcing_params,
                config,
                registered_variables,
            )

        # EVOLVE THE STATE
        if config.solver_mode == FINITE_VOLUME:
            primitive_state = _evolve_state_fv(
                primitive_state,
                dt,
                params.gamma,
                params.gravitational_constant,
                config,
                params,
                helper_data_pad,
                registered_variables,
            )
        elif config.solver_mode == FINITE_DIFFERENCE:
            primitive_state = _evolve_state_fd(
                primitive_state,
                dt,
                params.gamma,
                params.gravitational_constant,
                config,
                params,
                helper_data_pad,
                registered_variables,
            )

        time += dt

        # ----------------- ↑ CENTRAL UPDATE ↑ ----------------

        # If we are in the last time step, we also want to update the snapshot data.
        if config.use_specific_snapshot_timepoints and config.return_snapshots:
            snapshot_data = jax.lax.cond(
                jnp.abs(time - params.t_end) < 1e-12,
                update_snapshot_data,
                dont_update_snapshot_data,
                time,
                primitive_state,
                snapshot_data,
            )

        # progress bar update
        if config.progress_bar:
            jax.debug.callback(_show_progress, time, params.t_end)

        # packing the carry again
        if config.return_snapshots or config.activate_snapshot_callback:
            carry = (time, key, primitive_state, snapshot_data)
        else:
            carry = (time, key, primitive_state)

        return carry

    # -------------------------------------------------------------
    # ====================== ↑ Update step ↑ ======================
    # -------------------------------------------------------------

    # -------------------------------------------------------------
    # =================== ↓ loop-level logic ↓ ====================
    # -------------------------------------------------------------

    # Here we set up and start the actual time integration loops.
    # Depending on the configuration, this might be a fori loop
    # a while loop or a checkpointed while loop.

    def update_step_for(_, carry):
        return update_step(carry)

    def condition(carry):
        if config.return_snapshots or config.activate_snapshot_callback:
            t, _, _, _ = carry
        else:
            t, _, _ = carry
        return t < params.t_end

    if config.return_snapshots or config.activate_snapshot_callback:
        carry = (0.0, jax.random.key(42), primitive_state, snapshot_data)
    else:
        carry = (0.0, jax.random.key(42), primitive_state)

    if not config.fixed_timestep:
        if config.differentiation_mode == BACKWARDS:
            carry = checkpointed_while_loop(
                condition, update_step, carry, checkpoints=config.num_checkpoints
            )
        elif config.differentiation_mode == FORWARDS:
            carry = jax.lax.while_loop(condition, update_step, carry)
        else:
            raise ValueError("Unknown differentiation mode.")
    else:
        carry = jax.lax.fori_loop(0, config.num_timesteps, update_step_for, carry)

    # -------------------------------------------------------------
    # =================== ↑ loop-level logic ↑ ====================
    # -------------------------------------------------------------

    # -------------------------------------------------------------
    # ===================== ↓ return logic ↓ ======================
    # -------------------------------------------------------------

    # Finally, we need to unpack the results from the loops and
    # return them in the appropriate format.

    if config.return_snapshots or config.activate_snapshot_callback:
        _, _, primitive_state, snapshot_data = carry

        if config.return_snapshots:
            if config.snapshot_settings.return_final_state:
                if config.boundary_handling == GHOST_CELLS:
                    unpad_primitive_state = _unpad(primitive_state, config)
                else:
                    unpad_primitive_state = primitive_state
                
                snapshot_data = snapshot_data._replace(
                    final_state=unpad_primitive_state
                )
            return snapshot_data
        else:
            if config.boundary_handling == GHOST_CELLS:
                primitive_state = _unpad(primitive_state, config)
            if config.state_struct:
                return StateStruct(primitive_state=primitive_state)

            return primitive_state
    else:
        _, _, primitive_state = carry

        # unpad the primitive state if we padded it
        if config.boundary_handling == GHOST_CELLS:
            primitive_state = _unpad(primitive_state, config)

        if config.state_struct:
            return StateStruct(primitive_state=primitive_state)

        return primitive_state

    # -------------------------------------------------------------
    # ===================== ↑ return logic ↑ ======================
    # -------------------------------------------------------------


# general imports
from functools import partial
import jax.numpy as jnp
import jax

# typing
from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker
from typing import Union

# general astronomix
from astronomix._physics_modules._cosmic_rays.cr_fluid_equations import speed_of_sound_crs
from astronomix._finite_volume._riemann_solver._lax_friedrichs import _lax_friedrichs_solver
from astronomix._finite_volume._riemann_solver.hll import _am_hllc_solver, _hll_solver, _hllc_solver
from astronomix.variable_registry.registered_variables import RegisteredVariables

# fluid stuff
from astronomix._fluid_equations._equations import conserved_state_from_primitive, speed_of_sound
from astronomix._fluid_equations._fluxes import _euler_flux
from astronomix.option_classes.simulation_config import (
    AM_HLLC,
    HLL,
    HLLC,
    HLLC_LM,
    HYBRID_HLLC,
    LAX_FRIEDRICHS,
    STATE_TYPE,
    STATE_TYPE_ALTERED,
    SimulationConfig,
)


# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit, static_argnames=["config", "registered_variables", "flux_direction_index"]
)
def _riemann_solver(
    primitives_left: STATE_TYPE,
    primitives_right: STATE_TYPE,
    primitive_state: STATE_TYPE_ALTERED,
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    flux_direction_index: int,
) -> STATE_TYPE:
    """Wrapper function for the Riemann solver."""
    if config.riemann_solver == HLL:
        return _hll_solver(
            primitives_left,
            primitives_right,
            gamma,
            config,
            registered_variables,
            flux_direction_index,
        )
    elif config.riemann_solver == HLLC or config.riemann_solver == HLLC_LM:
        return _hllc_solver(
            primitives_left,
            primitives_right,
            gamma,
            config,
            registered_variables,
            flux_direction_index,
        )
    elif config.riemann_solver == AM_HLLC or config.riemann_solver == HYBRID_HLLC:
        return _am_hllc_solver(
            primitives_left,
            primitives_right,
            primitive_state,
            gamma,
            config,
            registered_variables,
            flux_direction_index,
        )
    elif config.riemann_solver == LAX_FRIEDRICHS:
        return _lax_friedrichs_solver(
            primitives_left,
            primitives_right,
            primitive_state,
            gamma,
            config,
            registered_variables,
            flux_direction_index,
        )

    else:
        raise ValueError("Riemann solver not supported.")



from types import NoneType
from typing import NamedTuple, Union

from jaxtyping import PyTree

class CNNMHDconfig(NamedTuple):
    cnn_mhd_corrector: bool = False
    network_static: Union[PyTree, NoneType] = None

class CNNMHDParams(NamedTuple):
    network_params: Union[PyTree, NoneType] = None


"""
The turbulent forcing module also draws from 
https://arxiv.org/pdf/2304.04360
"""

import jax
import jax.numpy as jnp
from functools import partial

from astronomix._physics_modules._turbulent_forcing._turbulent_forcing_options import TurbulentForcingParams
from astronomix.option_classes.simulation_config import SimulationConfig
from astronomix.variable_registry.registered_variables import RegisteredVariables

@partial(jax.jit, static_argnames=["config"])
def _create_forcing_field(
    key, 
    config: SimulationConfig,
):
    
    # in our case we assume a cubic box
    # with equal number of cells in each direction

    xsize = config.box_size
    ysize = config.box_size
    zsize = config.box_size

    nx = config.num_cells
    ny = config.num_cells
    nz = config.num_cells
    
    # wavenumbers using fftfreq
    kx = 2.0 * jnp.pi * jnp.fft.fftfreq(nx, d=xsize/nx)
    ky = 2.0 * jnp.pi * jnp.fft.fftfreq(ny, d=ysize/ny)
    kz = 2.0 * jnp.pi * jnp.fft.fftfreq(nz, d=zsize/nz)
    
    # broadcasting instead of meshgrid
    kx_3d = kx.reshape(nx, 1, 1)
    ky_3d = ky.reshape(1, ny, 1)
    kz_3d = kz.reshape(1, 1, nz)
    
    k_squared = kx_3d**2 + ky_3d**2 + kz_3d**2
    kk = jnp.sqrt(k_squared)
    
    # power spectrum of the forcing
    kpk = 4.0 * jnp.pi / config.box_size
    Pk = kk**6 * jnp.exp(-8.0 * kk / kpk)
    
    key, sk1, sk2 = jax.random.split(key, 3)

    raw_noise = jax.random.normal(sk1, shape=(3, nx, ny, nz)) + \
                1j * jax.random.normal(sk2, shape=(3, nx, ny, nz))

    cwx = jnp.sqrt(Pk) * raw_noise[0]
    cwy = jnp.sqrt(Pk) * raw_noise[1]
    cwz = jnp.sqrt(Pk) * raw_noise[2]
    
    # DC mode to zero
    cwx = cwx.at[0, 0, 0].set(0.0 + 0.0j)
    cwy = cwy.at[0, 0, 0].set(0.0 + 0.0j)
    cwz = cwz.at[0, 0, 0].set(0.0 + 0.0j)
    
    # project out compressible component
    k_squared_safe = jnp.where(k_squared == 0.0, 1.0, k_squared)
    div_k = (kx_3d * cwx + ky_3d * cwy + kz_3d * cwz) / k_squared_safe
    div_k = div_k.at[0, 0, 0].set(0.0 + 0.0j)
    cwx = cwx - kx_3d * div_k
    cwy = cwy - ky_3d * div_k
    cwz = cwz - kz_3d * div_k

    # get real space fields
    wx_real = jnp.real(jnp.fft.ifftn(cwx))
    wy_real = jnp.real(jnp.fft.ifftn(cwy))
    wz_real = jnp.real(jnp.fft.ifftn(cwz))
    
    return key, wx_real, wy_real, wz_real

@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _apply_forcing(
    key,
    primitive_state,
    dt,
    turbulent_forcing_params: TurbulentForcingParams,
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
):
    
    key, wx_real, wy_real, wz_real = _create_forcing_field(key, config)

    Edot = turbulent_forcing_params.energy_injection_rate
    dtforc = dt
    dV = config.grid_spacing**3
    
    # get density and velocities
    rho = primitive_state[registered_variables.density_index]
    u = primitive_state[registered_variables.velocity_index.x]
    v = primitive_state[registered_variables.velocity_index.y]
    w = primitive_state[registered_variables.velocity_index.z]

    # compute terms for quadratic equation
    # a * amp^2 + b * amp + c = 0
    tempa = 0.5 * jnp.sum(rho * (wx_real**2 + wy_real**2 + wz_real**2))
    tempb = jnp.sum(rho * u * wx_real + rho * v * wy_real + rho * w * wz_real)
    tempc = -Edot * dtforc / dV
    
    # solve quadratic equation for amplitude
    discriminant = tempb**2 - 4.0 * tempa * tempc
    
    # conditional to handle edge cases
    amp = jax.lax.cond(
        (discriminant >= 0) & (jnp.abs(tempa) > 1e-10),
        lambda: (-tempb + jnp.sqrt(discriminant)) / (2.0 * tempa),
        lambda: 0.0
    )
    
    # apply forcing to momentum
    primitive_state = primitive_state.at[registered_variables.velocity_index.x].add(
        amp * wx_real
    )
    primitive_state = primitive_state.at[registered_variables.velocity_index.y].add(
        amp * wy_real
    )
    primitive_state = primitive_state.at[registered_variables.velocity_index.z].add(
        amp * wz_real
    )
    return key, primitive_state


# @partial(jax.jit, static_argnames=["config", "registered_variables"])
# def apply_forcing(
#     key,
#     conserved_state,
#     dt,
#     turbulent_forcing_params: TurbulentForcingParams,
#     config: SimulationConfig,
#     registered_variables: RegisteredVariables,
# ):
    
#     key, wx_real, wy_real, wz_real = create_forcing_field(key, config)

#     Edot = turbulent_forcing_params.energy_injection_rate
#     dtforc = dt
#     dV = config.grid_spacing**3
    
#     # get density and velocities
#     rho = conserved_state[registered_variables.density_index]
#     rhou = conserved_state[registered_variables.momentum_index.x]
#     rhov = conserved_state[registered_variables.momentum_index.y]
#     rhow = conserved_state[registered_variables.momentum_index.z]

#     # compute terms for quadratic equation
#     # a * amp^2 + b * amp + c = 0
#     tempa = 0.5 * jnp.sum(rho * (wx_real**2 + wy_real**2 + wz_real**2))
#     tempb = jnp.sum(rhou * wx_real + rhov * wy_real + rhow * wz_real)
#     tempc = -Edot * dtforc / dV
    
#     # solve quadratic equation for amplitude
#     discriminant = tempb**2 - 4.0 * tempa * tempc
    
#     # conditional to handle edge cases
#     amp = jax.lax.cond(
#         (discriminant >= 0) & (jnp.abs(tempa) > 1e-10),
#         lambda: (-tempb + jnp.sqrt(discriminant)) / (2.0 * tempa),
#         lambda: 0.0
#     )
    
#     # apply forcing to momentum
#     conserved_state = conserved_state.at[registered_variables.momentum_index.x].add(
#         rho * amp * wx_real
#     )
#     conserved_state = conserved_state.at[registered_variables.momentum_index.y].add(
#         rho * amp * wy_real
#     )
#     conserved_state = conserved_state.at[registered_variables.momentum_index.z].add(
#         rho * amp * wz_real
#     )

#     # update the total energy
#     # kinetic energy = 0.5 * rho * u_squared
#     # where rho u -> rho (u + amp w)
#     # -> change in kinetic energy = 0.5 * rho * [ (u + amp w)^2 - u^2 ]
#     new_squared_velocity = (
#         (rhou / rho + amp * wx_real) ** 2
#         + (rhov / rho + amp * wy_real) ** 2
#         + (rhow / rho + amp * wz_real) ** 2
#     )
#     old_squared_velocity = (
#         (rhou / rho) ** 2
#         + (rhov / rho) ** 2
#         + (rhow / rho) ** 2
#     )
#     delta_kinetic_energy = 0.5 * rho * (new_squared_velocity - old_squared_velocity)
#     conserved_state = conserved_state.at[registered_variables.energy_index].add(delta_kinetic_energy)
    
#     return key, conserved_state


from typing import NamedTuple
import jax.numpy as jnp

from astronomix._physics_modules._cnn_mhd_corrector._cnn_mhd_corrector_options import CNNMHDconfig
from astronomix._physics_modules._cooling.cooling_options import CoolingParams
from astronomix._physics_modules._cosmic_rays.cosmic_ray_options import CosmicRayParams
from astronomix._physics_modules._neural_net_force._neural_net_force_options import NeuralNetForceParams
from astronomix._physics_modules._stellar_wind.stellar_wind_options import WindParams
from astronomix._physics_modules._turbulent_forcing._turbulent_forcing_options import TurbulentForcingParams

class SimulationParams(NamedTuple):
    """
    Different from the simulation configuration, the simulation parameters
    do not require recompilation when changed. The simulation can be 
    differentiated with respect to them.
    """

    #: The Courant-Friedrichs-Lewy number, a factor
    #: in the time step calculation.
    C_cfl: float = 0.4

    #: Gravitational constant.
    gravitational_constant: float = 1.0

    #: The adiabatic index of the gas.
    gamma: float = 5/3

    #: Minimum allowed density.
    #: NOTE: CURRENTLY ONLY USED IN 
    #: FINITE DIFFERENCE MODE IF
    #: config.enforce_positivity IS TRUE.
    minimum_density: float = 1e-14

    #: Minimum allowed pressure.
    #: NOTE: CURRENTLY ONLY USED IN 
    #: FINITE DIFFERENCE MODE IF
    #: config.enforce_positivity IS TRUE.
    minimum_pressure: float = 1e-14

    #: The maximum time step.
    dt_max: float = 0.001

    #: The final time of the simulation.
    t_end: float = 0.2

    #: Snapshot timepoints
    snapshot_timepoints: jnp.array = jnp.array([0.0])

    # parameters of physics modules

    #: The parameters of the turbulent forcing module.
    turbulent_forcing_params: TurbulentForcingParams = TurbulentForcingParams()

    #: The parameters of the stellar wind module.
    wind_params: WindParams = WindParams()

    #: Cosmic ray parameters
    cosmic_ray_params: CosmicRayParams = CosmicRayParams()

    #: The parameters of the cooling module.
    cooling_params: CoolingParams = CoolingParams()

    #: The parameters of the neural network force module.
    neural_net_force_params: NeuralNetForceParams = NeuralNetForceParams()

    #: The parameters of the CNN MHD corrector module.
    cnn_mhd_corrector_params: CNNMHDconfig = CNNMHDconfig()


# general imports
import jax.numpy as jnp
import jax
from functools import partial


# type checking imports
from jaxtyping import Array, Float
from beartype import beartype as typechecker
from typing import Union

# general astronomix imports
from astronomix._finite_difference._fluid_equations._eigen import _eigen_all_lambdas
from astronomix._finite_difference._fluid_equations._equations import conserved_state_from_primitive_mhd, primitive_state_from_conserved_mhd
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import (
    STATE_TYPE,
    SimulationConfig,
)

from astronomix.option_classes.simulation_params import SimulationParams


@partial(jax.jit, static_argnames=["config", "registered_variables"])
def _cfl_time_step_fd(
    primitive_state: STATE_TYPE,
    grid_spacing: Union[float, Float[Array, ""]],
    dt_max: Union[float, Float[Array, ""]],
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    params: SimulationParams,
    registered_variables: RegisteredVariables,
    C_CFL: Union[float, Float[Array, ""]] = 0.8,
) -> Float[Array, ""]:
    
    # TODO: use specific lambda function

    conserved_state = conserved_state_from_primitive_mhd(
        primitive_state, gamma, registered_variables
    )

    lambda_x = _eigen_all_lambdas(
        conserved_state, params.minimum_density, params.minimum_pressure, gamma, registered_variables
    )

    lambda_x = jnp.max(jnp.abs(lambda_x))

    qy = jnp.transpose(conserved_state, (0, 2, 1, 3))
    momentum_x = qy[registered_variables.momentum_index.x]
    momentum_y = qy[registered_variables.momentum_index.y]
    B_x = qy[registered_variables.magnetic_index.x]
    B_y = qy[registered_variables.magnetic_index.y]
    qy = qy.at[registered_variables.momentum_index.x].set(momentum_y)
    qy = qy.at[registered_variables.momentum_index.y].set(momentum_x)
    qy = qy.at[registered_variables.magnetic_index.x].set(B_y)
    qy = qy.at[registered_variables.magnetic_index.y].set(B_x)

    # lambda_y, _, _ = _eigen_x(
    #     qy, gamma, registered_variables
    # )

    lambda_y = _eigen_all_lambdas(
        qy, params.minimum_density, params.minimum_pressure, gamma, registered_variables
    )

    lambda_y = jnp.max(jnp.abs(lambda_y))

    qz = jnp.transpose(conserved_state, (0, 3, 2, 1))
    momentum_x = qz[registered_variables.momentum_index.x]
    momentum_z = qz[registered_variables.momentum_index.z]
    B_x = qz[registered_variables.magnetic_index.x]
    B_z = qz[registered_variables.magnetic_index.z]
    qz = qz.at[registered_variables.momentum_index.x].set(momentum_z)
    qz = qz.at[registered_variables.momentum_index.z].set(momentum_x)
    qz = qz.at[registered_variables.magnetic_index.x].set(B_z)
    qz = qz.at[registered_variables.magnetic_index.z].set(B_x)

    # lambda_z, _, _ = _eigen_x(
    #     qz, gamma, registered_variables
    # )

    lambda_z = _eigen_all_lambdas(
        qz, params.minimum_density, params.minimum_pressure, gamma, registered_variables
    )

    lambda_z = jnp.max(jnp.abs(lambda_z))

    dt_cfl = C_CFL * grid_spacing / (lambda_x + lambda_y + lambda_z)
    dt_cfl = jnp.minimum(dt_cfl, dt_max)

    return dt_cfl


import jax.numpy as jnp
import jax

from astronomix._stencil_operations._stencil_operations import _stencil_add
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.option_classes.simulation_config import (
    STATE_TYPE,
    STATE_TYPE_ALTERED,
    SimulationConfig,
)


@jax.jit
def _minmod(a, b):
    return 0.5 * (jnp.sign(a) + jnp.sign(b)) * jnp.minimum(jnp.abs(a), jnp.abs(b))


def _minmod3(a, b, c):
    """Minmod function for three arguments in JAX."""
    same_sign = (jnp.sign(a) == jnp.sign(b)) & (jnp.sign(b) == jnp.sign(c))
    return jnp.where(
        same_sign,
        jnp.sign(a) * jnp.minimum(jnp.abs(a), jnp.minimum(jnp.abs(b), jnp.abs(c))),
        0.0,
    )


def _double_minmod(a, b):
    """
    Double minmod limiter.
    """
    return jnp.where(a * b > 0, _minmod3((a + b) / 2, 2 * a, 2 * b), 0.0)


@jax.jit
def _maxmod(a, b):
    return 0.5 * (jnp.sign(a) + jnp.sign(b)) * jnp.maximum(jnp.abs(a), jnp.abs(b))


def _superbee(a, b):
    """
    Superbee limiter.
    """
    return jnp.where(a * b > 0, _minmod(_maxmod(a, b), _minmod(2 * a, 2 * b)), 0.0)


# TODO: bring into common interface
def _van_albada_limiter(
    backward_difference: STATE_TYPE,
    forward_difference: STATE_TYPE,
    config: SimulationConfig,
) -> STATE_TYPE_ALTERED:
    """
    van Albada limited gradients along an axis
    """

    grid_spacing = config.grid_spacing
    epsilon = 3 * grid_spacing
    limited_gradients = (
        (forward_difference**2 + epsilon) * backward_difference
        + (backward_difference**2 + epsilon) * forward_difference
    ) / (forward_difference**2 + backward_difference**2 + 2 * epsilon)

    return limited_gradients



from typing import NamedTuple

class CosmicRayConfig(NamedTuple):

    #: main switch for cosmic rays
    cosmic_rays: bool = False

    #: turn on injection of CRs at shocks
    diffusive_shock_acceleration: bool = False

class CosmicRayParams(NamedTuple):

    #: starting time of diffusive shock acceleration
    diffusive_shock_acceleration_start_time: float = 0.0

    #: efficiency of diffusive shock acceleration
    diffusive_shock_acceleration_efficiency: float = 0.1


"""
Equations for 3D adiabatic ideal magnetohydrodynamics (MHD).
"""

import jax.numpy as jnp
import jax
from functools import partial

from jaxtyping import Array, Float

from typing import Union

from astronomix.variable_registry.registered_variables import RegisteredVariables
from astronomix.option_classes.simulation_config import (
    FIELD_TYPE,
    STATE_TYPE,
    SimulationConfig,
)


@partial(jax.jit, static_argnames=["registered_variables"])
def _u_squared3D(
    primitive_state: STATE_TYPE,
    registered_variables: RegisteredVariables,
) -> FIELD_TYPE:
    return (
        primitive_state[registered_variables.velocity_index.x] ** 2
        + primitive_state[registered_variables.velocity_index.y] ** 2
        + primitive_state[registered_variables.velocity_index.z] ** 2
    )


def _b_squared3D(
    primitive_state: STATE_TYPE,
    registered_variables: RegisteredVariables,
) -> FIELD_TYPE:
    return (
        primitive_state[registered_variables.magnetic_index.x] ** 2
        + primitive_state[registered_variables.magnetic_index.y] ** 2
        + primitive_state[registered_variables.magnetic_index.z] ** 2
    )


@jax.jit
def thermal_pressure_from_energy_mhd(E, rho, u_squared, b_squared, gamma):
    """Calculate the pressure from the total energy in MHD.

    Args:
        E: The total energy.
        rho: The density.
        u_squared: The squared velocity.
        b_squared: The squared magnetic field.
        gamma: The adiabatic index.
    Returns:
        The pressure.
    """
    return (gamma - 1) * (E - 0.5 * rho * u_squared - 0.5 * b_squared)


@jax.jit
def total_energy_from_primitives_mhd(rho, u_squared, p, b_squared, gamma):
    return p / (gamma - 1) + 0.5 * rho * u_squared + 0.5 * b_squared


@partial(jax.jit, static_argnames=["registered_variables"])
def conserved_state_from_primitive_mhd(
    primitive_state: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    """
    For now only 3D.
    """

    rho = primitive_state[registered_variables.density_index]

    u_squared = _u_squared3D(primitive_state, registered_variables)

    p = primitive_state[registered_variables.pressure_index]  # thermal pressure

    b_squared = _b_squared3D(primitive_state, registered_variables)

    # calculate total energy
    E = total_energy_from_primitives_mhd(rho, u_squared, p, b_squared, gamma)

    # create conserved state
    conserved_state = primitive_state.at[registered_variables.pressure_index].set(E)

    # set momentum density
    conserved_state = conserved_state.at[
        registered_variables.velocity_index.x : registered_variables.velocity_index.z
        + 1
    ].set(
        rho
        * primitive_state[
            registered_variables.velocity_index.x : registered_variables.velocity_index.z
            + 1
        ]
    )

    return conserved_state


@partial(jax.jit, static_argnames=["registered_variables", 'config'])
def primitive_state_from_conserved_mhd(
    conserved_state: STATE_TYPE,
    rhomin: Union[float, Float[Array, ""]],
    pgmin: Union[float, Float[Array, ""]],
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
) -> STATE_TYPE:
    """
    For now only 3D.
    """

    rho = conserved_state[registered_variables.density_index]
    E = conserved_state[registered_variables.pressure_index]

    ux = conserved_state[registered_variables.velocity_index.x] / rho
    uy = conserved_state[registered_variables.velocity_index.y] / rho
    uz = conserved_state[registered_variables.velocity_index.z] / rho

    u_squared = ux**2 + uy**2 + uz**2

    b_squared = _b_squared3D(conserved_state, registered_variables)

    p = thermal_pressure_from_energy_mhd(E, rho, u_squared, b_squared, gamma)

    # set the primitive state

    # pressure
    primitive_state = conserved_state.at[registered_variables.pressure_index].set(p)

    # velocities
    primitive_state = primitive_state.at[registered_variables.velocity_index.x].set(ux)
    primitive_state = primitive_state.at[registered_variables.velocity_index.y].set(uy)
    primitive_state = primitive_state.at[registered_variables.velocity_index.z].set(uz)

    if config.enforce_positivity:
        # enforce positivity of density and pressure
        primitive_state = primitive_state.at[registered_variables.density_index].set(
            jnp.maximum(
                primitive_state[registered_variables.density_index], rhomin
            )
        )
        primitive_state = primitive_state.at[registered_variables.pressure_index].set(
            jnp.maximum(
                primitive_state[registered_variables.pressure_index], pgmin
            )
        )

    return primitive_state


@partial(jax.jit, static_argnames=["registered_variables"])
def total_pressure_from_conserved_mhd(
    conserved_state: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    registered_variables: RegisteredVariables,
) -> FIELD_TYPE:
    """
    For now only 3D.
    """

    rho = conserved_state[registered_variables.density_index]
    E = conserved_state[registered_variables.pressure_index]

    ux = conserved_state[registered_variables.velocity_index.x] / rho
    uy = conserved_state[registered_variables.velocity_index.y] / rho
    uz = conserved_state[registered_variables.velocity_index.z] / rho

    u_squared = ux**2 + uy**2 + uz**2

    b_squared = _b_squared3D(conserved_state, registered_variables)

    p_thermal = thermal_pressure_from_energy_mhd(E, rho, u_squared, b_squared, gamma)

    return p_thermal + 0.5 * b_squared



# general imports
from functools import partial
import jax
import jax.numpy as jnp

# typechecking imports
from beartype import beartype as typechecker
from jaxtyping import jaxtyped
from typing import Union

# general astronomix imports
from astronomix._stencil_operations._stencil_operations import _stencil_add
from astronomix.data_classes.simulation_helper_data import HelperData
from astronomix.option_classes.simulation_config import CARTESIAN, MINMOD, OSHER, DOUBLE_MINMOD, SUPERBEE, SPHERICAL, STATE_TYPE, STATE_TYPE_ALTERED, VAN_ALBADA, VAN_ALBADA_PP, SimulationConfig

# limiter imports
from astronomix._finite_volume._state_evolution.limiters import _double_minmod, _minmod, _superbee, _van_albada_limiter


# @jaxtyped(typechecker=typechecker)
@partial(jax.jit, static_argnames=['config', 'axis'])
def _calculate_limited_gradients(
    primitive_state: STATE_TYPE,
    config: SimulationConfig,
    helper_data: HelperData,
    axis: int
) -> STATE_TYPE_ALTERED:

    # We first need to calculate the distances between the cells.
    # For 1D simulations in spherical geometry, we have to mind
    # that the distances of the volumetric centers to the cell 
    # interfaces are not equal.

    if config.geometry == CARTESIAN:
        cell_distances_left = config.grid_spacing # distances r_i - r_{i-1}
        cell_distances_right = config.grid_spacing # distances r_{i+1} - r_i
    elif config.geometry == SPHERICAL and config.dimensionality == 1:
        # calculate the distances
        cell_distances_left = _stencil_add(helper_data.volumetric_centers, indices = (0, -1), factors = (1.0, -1.0), axis = 0)
        cell_distances_right = _stencil_add(helper_data.volumetric_centers, indices = (1, 0), factors = (1.0, -1.0), axis = 0)
    else:
        raise ValueError("Geometry and dimensionality combination not supported.")

    
    # Next we calculate the finite differences of consecutive cells.
    # a is the left difference, b the right difference for cells
    # 1 to num_cells - 1.

    # backward
    backward_diff = _stencil_add(primitive_state, indices = (0, -1), factors = (1.0, -1.0), axis = axis) / cell_distances_left

    # forward
    forward_diff = _stencil_add(primitive_state, indices = (1, 0), factors = (1.0, -1.0), axis = axis) / cell_distances_right

    # We apply limiting to not create new extrema in regions where consecutive finite
    # differences differ strongly.

    if config.limiter == MINMOD:
        # Limited average formulations:
        limited_gradients = _minmod(backward_diff, forward_diff)
    elif config.limiter == SUPERBEE:
        limited_gradients = _superbee(backward_diff, forward_diff)
    elif config.limiter == DOUBLE_MINMOD:
        limited_gradients = _double_minmod(backward_diff, forward_diff)
    elif config.limiter == VAN_ALBADA or config.limiter == VAN_ALBADA_PP:
        # van Albada limiter
        limited_gradients = _van_albada_limiter(
            backward_diff,
            forward_diff,
            config
        )
    elif config.limiter == OSHER:
        # Quotient formulation:
        epsilon = 1e-11  # Small constant to prevent division by zero
        g = jnp.where(
            jnp.abs(backward_diff) > epsilon,  # Avoid division if `a` is very small
            forward_diff / (backward_diff + epsilon),  # Add epsilon to `a` for numerical stability
            jnp.zeros_like(backward_diff)
        )
        # slope_limited = jnp.maximum(0, jnp.minimum(1, g))  # Minmod limiter
        slope_limited = jnp.maximum(0, jnp.minimum(1.3, g))  # Osher limiter with beta = 1.3
        # ospre limiter
        # slope_limited = (1.5 * (g ** 2 + g)) / (g ** 2 + g + 1)
        limited_gradients = slope_limited * backward_diff
    else:
        raise ValueError("Unknown limiter.")

    return limited_gradients


# general imports
from functools import partial
import jax.numpy as jnp
import jax

# typing
from jaxtyping import Array, Float, jaxtyped
from beartype import beartype as typechecker
from typing import Union

# general astronomix
from astronomix._physics_modules._cosmic_rays.cr_fluid_equations import speed_of_sound_crs
from astronomix._stencil_operations._stencil_operations import _stencil_add
from astronomix.variable_registry.registered_variables import RegisteredVariables

# fluid stuff
from astronomix._fluid_equations._equations import (
    conserved_state_from_primitive,
    get_absolute_velocity,
    speed_of_sound,
)
from astronomix._fluid_equations._fluxes import _euler_flux
from astronomix.option_classes.simulation_config import (
    AM_HLLC,
    HLLC_LM,
    HYBRID_HLLC,
    STATE_TYPE,
    SimulationConfig,
)


# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit, static_argnames=["config", "registered_variables", "flux_direction_index"]
)
def _hll_solver(
    primitives_left: STATE_TYPE,
    primitives_right: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    flux_direction_index: int,
) -> STATE_TYPE:
    """
    Returns the conservative fluxes.

    Args:
        primitives_left: States left of the interfaces.
        primitives_right: States right of the interfaces.
        gamma: The adiabatic index.

    Returns:
        The conservative fluxes at the interfaces.

    """

    rho_L = primitives_left[registered_variables.density_index]

    u_L = primitives_left[flux_direction_index]

    rho_R = primitives_right[registered_variables.density_index]

    u_R = primitives_right[flux_direction_index]

    p_L = primitives_left[registered_variables.pressure_index]
    p_R = primitives_right[registered_variables.pressure_index]

    # calculate the sound speeds
    if not config.cosmic_ray_config.cosmic_rays:
        c_L = speed_of_sound(rho_L, p_L, gamma)
        c_R = speed_of_sound(rho_R, p_R, gamma)
    else:
        c_L = speed_of_sound_crs(primitives_left, registered_variables)
        c_R = speed_of_sound_crs(primitives_right, registered_variables)

    # get the left and right states and fluxes
    fluxes_left = _euler_flux(
        primitives_left, gamma, config, registered_variables, flux_direction_index
    )
    fluxes_right = _euler_flux(
        primitives_right, gamma, config, registered_variables, flux_direction_index
    )

    # very simple approach for the wave velocities
    wave_speeds_right_plus = jnp.maximum(jnp.maximum(u_L + c_L, u_R + c_R), 0)
    wave_speeds_left_minus = jnp.minimum(jnp.minimum(u_L - c_L, u_R - c_R), 0)

    # wave_speeds_right_plus = jnp.maximum(u_L - c_L, 0)
    # wave_speeds_left_minus = jnp.minimum(u_R + c_R, 0)

    # get the left and right conserved variables
    conserved_left = conserved_state_from_primitive(
        primitives_left, gamma, config, registered_variables
    )
    conserved_right = conserved_state_from_primitive(
        primitives_right, gamma, config, registered_variables
    )

    # calculate the interface HLL fluxes
    # F = (S_R * F_L - S_L * F_R + S_L * S_R * (U_R - U_L)) / (S_R - S_L)
    fluxes = (
        wave_speeds_right_plus * fluxes_left
        - wave_speeds_left_minus * fluxes_right
        + wave_speeds_left_minus
        * wave_speeds_right_plus
        * (conserved_right - conserved_left)
    ) / (wave_speeds_right_plus - wave_speeds_left_minus)

    return fluxes


# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit,
    static_argnames=[
        "config",
        "registered_variables",
        "flux_direction_index",
        "hllc_lm",
        "low_mach_dissipation_control",
    ],
)
def _hllc_solver(
    primitives_left: STATE_TYPE,
    primitives_right: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    flux_direction_index: int,
    hllc_lm: bool = False,
    low_mach_dissipation_control: bool = False,
) -> STATE_TYPE:

    # There seem to be problems for 1d radial, maybe because the same stuff that has to be
    # taken into consideration in the general scheme (averaging based on density might be
    # problematic, as the surface increases radially, ...) is not taken into account here.
    # Maybe interesting for future research, for now HLL works fine.

    rho_L = primitives_left[registered_variables.density_index]
    u_L = primitives_left[flux_direction_index]

    rho_R = primitives_right[registered_variables.density_index]
    u_R = primitives_right[flux_direction_index]

    p_L = primitives_left[registered_variables.pressure_index]
    p_R = primitives_right[registered_variables.pressure_index]

    # calculate the sound speeds
    if not config.cosmic_ray_config.cosmic_rays:
        c_L = speed_of_sound(rho_L, p_L, gamma)
        c_R = speed_of_sound(rho_R, p_R, gamma)
    else:
        c_L = speed_of_sound_crs(primitives_left, registered_variables)
        c_R = speed_of_sound_crs(primitives_right, registered_variables)

    # get the left and right states and fluxes
    F_L = _euler_flux(
        primitives_left, gamma, config, registered_variables, flux_direction_index
    )
    F_R = _euler_flux(
        primitives_right, gamma, config, registered_variables, flux_direction_index
    )

    # Roe average of the velocity
    u_hat = (jnp.sqrt(rho_L) * u_L + jnp.sqrt(rho_R) * u_R) / (
        jnp.sqrt(rho_L) + jnp.sqrt(rho_R)
    )

    # Roe average of the sound speed
    c_hat_squared = (c_L**2 * jnp.sqrt(rho_L) + c_R**2 * jnp.sqrt(rho_R)) / (
        jnp.sqrt(rho_L) + jnp.sqrt(rho_R)
    ) + 0.5 * (
        jnp.sqrt(rho_L) * jnp.sqrt(rho_R) / (jnp.sqrt(rho_L) + jnp.sqrt(rho_R)) ** 2
    ) * (u_R - u_L) ** 2
    c_hat = jnp.sqrt(c_hat_squared)

    # Einfeldt estimates of maximum left and right signal speeds
    S_L = jnp.minimum(u_L - c_L, u_hat - c_hat)
    S_R = jnp.maximum(u_R + c_R, u_hat + c_hat)

    # contact wave signal speed
    S_star = (p_R - p_L + rho_L * u_L * (S_L - u_L) - rho_R * u_R * (S_R - u_R)) / (
        rho_L * (S_L - u_L) - rho_R * (S_R - u_R)
    )

    # intermediate states
    U_L = conserved_state_from_primitive(
        primitives_left, gamma, config, registered_variables
    )
    U_R = conserved_state_from_primitive(
        primitives_right, gamma, config, registered_variables
    )

    U_star_L = U_L.at[flux_direction_index].set(rho_L * S_star)
    U_star_L = U_star_L.at[registered_variables.pressure_index].add(
        (S_star - u_L) * (rho_L * S_star + p_L / (S_L - u_L))
    )
    U_star_L = U_star_L * (S_L - u_L) / (S_L - S_star)

    U_star_R = U_R.at[flux_direction_index].set(rho_R * S_star)
    U_star_R = U_star_R.at[registered_variables.pressure_index].add(
        (S_star - u_R) * (rho_R * S_star + p_R / (S_R - u_R))
    )
    U_star_R = U_star_R * (S_R - u_R) / (S_R - S_star)

    # HLLC-LM adaptation
    # following
    # https://doi.org/10.1016/j.jcp.2020.109762
    if config.riemann_solver == HLLC_LM or hllc_lm:
        Ma_limit = 0.1
        Ma_local = jnp.maximum(jnp.abs(u_L / c_L), jnp.abs(u_R / c_R))
        phi = jnp.sin(jnp.minimum(1, Ma_local / Ma_limit) * jnp.pi / 2)
        S_Llm = S_L * phi
        S_Rlm = S_R * phi

    if config.riemann_solver == HLLC_LM or hllc_lm:
        S_Lstar = S_Llm
        S_Rstar = S_Rlm
    else:
        S_Lstar = S_L
        S_Rstar = S_R

    bulk_flux_star = 0.5 * (F_L + F_R)
    dissipation_term_star = 0.5 * (
        S_Lstar * (U_star_L - U_L)
        + jnp.abs(S_star) * (U_star_L - U_star_R)
        + S_Rstar * (U_star_R - U_R)
    )

    if low_mach_dissipation_control:
        absolute_velocity_L = get_absolute_velocity(
            primitives_left, config, registered_variables
        )
        absolute_velocity_R = get_absolute_velocity(
            primitives_right, config, registered_variables
        )
        Ma_tilde = jnp.maximum(
            jnp.abs(absolute_velocity_L / c_L), jnp.abs(absolute_velocity_R / c_R)
        )
        f = jnp.minimum(1, Ma_tilde)

        if config.dimensionality == 1:
            velocity_start_index = registered_variables.velocity_index
        else:
            velocity_start_index = registered_variables.velocity_index.x

        dissipation_term_star = dissipation_term_star.at[
            velocity_start_index : velocity_start_index + config.dimensionality
        ].set(
            f
            * dissipation_term_star[
                velocity_start_index : velocity_start_index + config.dimensionality
            ]
        )

    F_star = bulk_flux_star + dissipation_term_star

    fluxes = jnp.where(S_L >= 0, F_L, F_star)
    fluxes = jnp.where(S_R <= 0, F_R, fluxes)

    return fluxes


# @jaxtyped(typechecker=typechecker)
@partial(
    jax.jit, static_argnames=["config", "registered_variables", "flux_direction_index"]
)
def _am_hllc_solver(
    primitives_left: STATE_TYPE,
    primitives_right: STATE_TYPE,
    primitive_state: STATE_TYPE,
    gamma: Union[float, Float[Array, ""]],
    config: SimulationConfig,
    registered_variables: RegisteredVariables,
    flux_direction_index: int,
) -> STATE_TYPE:
    """
    https://www.sciencedirect.com/science/article/pii/S1007570425005891
    """

    d = config.grid_spacing
    C_th = 0.05
    a = speed_of_sound(
        primitive_state[registered_variables.density_index],
        primitive_state[registered_variables.pressure_index],
        gamma,
    )
    # not optimal, sum over all dimensions
    # is carried out once per dimension
    div_v = sum(
        _stencil_add(
            primitive_state[i + 1], indices=(1, -1), factors=(1.0, -1.0), axis=i
        )
        / (2 * d)
        for i in range(config.dimensionality)
    )
    g = jnp.where(div_v < -C_th * a / d, 1, 0)

    if config.riemann_solver == AM_HLLC:
        low_mach_dissipation_control = True
    elif config.riemann_solver == HYBRID_HLLC:
        low_mach_dissipation_control = False
    else:
        raise ValueError("Riemann solver not supported for AM-HLLC.")

    fluxes_hllc = _hllc_solver(
        primitives_left,
        primitives_right,
        gamma,
        config,
        registered_variables,
        flux_direction_index,
        low_mach_dissipation_control=low_mach_dissipation_control,
    )
    fluxes_hllc_lm = _hllc_solver(
        primitives_left,
        primitives_right,
        gamma,
        config,
        registered_variables,
        flux_direction_index,
        hllc_lm=True,
        low_mach_dissipation_control=low_mach_dissipation_control,
    )

    return g * fluxes_hllc_lm + (1 - g) * fluxes_hllc
